
crisisList <- function
(metricsTS,  # Multivariate time series of the data,
             # as a data frame with the columns
             # corresponding to the metrics and KPIs and the
             # the rows corresponding to the time
             # periods.
 kpiInd,     # The indices of the KPIs, i.e. the column indexes
             # in metricsTS that correspond to the KPIs.
             # Should be a
             # vector of length <= the number of
             # columns 
 kpiCutoff,  # A vector of the thresholds for each of the
             # KPIs that define a crisis.  Should have
             # the same length as kpiInd.
 minLength = 1,    # The minimum crisis length
 crisCutoff = 1e4,    # The max number of time periods to include
                      # after the start of a crisis.
 preCrisis = 0,       # The num. of time periods to include just
                      # before the start of each crisis.
 minGap = 2           # The minimum gap between to be considered separate crises
 ){

  nData <- dim( metricsTS )[1]
  nKPIs <- length( kpiInd )
  nMetrics <- dim( metricsTS )[2]
  if( nKPIs > nMetrics )
    stop("The number of KPIs is greater than the number of columns of metricsTS")

  # Calculate a vector of crisis indicators for each time period.
  isCrisis <- rep( F, nData )
  for( i in 1:nKPIs ){
    thisKPI <- kpiInd[i]
    isCrisis <- isCrisis | ( metricsTS[,thisKPI] > kpiCutoff[i] )
  }

  ### Convert metricsTS by removing the KPI columns
  ### and thresholding the remaining metrics according
  ### to "cold", "regular", and "hot"

  # Calculate the definitions of "hot" and "cold" for each metric.
  # Use only the non-crisis data.
  hotColdThresh <- matrix(0.0, nrow = nMetrics, ncol = 2)
  for( i in 1:nMetrics ){
    hotColdThresh[i,] <- quantile( metricsTS[!isCrisis,i], c(0.02, 0.98) );
  }

  # threshold the metrics
  metricsThresh <- metricsTS
  for( i in 1:nMetrics ){
    metricsThresh[,i] <- (metricsThresh[,i] > hotColdThresh[i,1]) + (metricsThresh[,i] > hotColdThresh[i,2])
  }
  
  # Remove the KPI columns
  KPI <- rep(F, nMetrics)
  KPI[ kpiInd ] <- T
  metricsThresh <- metricsThresh[,!KPI, drop = F]

#  print("Time series observations formatted.  The correlations between the different metrics are:")
#  print( cor( metricsThresh ))
  
  ### Obtain a list of the crises, containing data frames 
  ### of the metrics during the crisis periods.

  # The rle function finds the repetitive sequences of
  # TRUE or FALSE in the isCrisis vector.
  crisRle <- rle( isCrisis )
  # find the number of potential crises in the data
  # (all sequences of TRUE values)
  nCris <- sum( crisRle$values == T )

  # Iterate through the potential crises to find
  # those that satisfy the conditions, and store
  # the thresholded non-KPI data for those crises
  # in "crisisDat"
  currentInd <- 1
  potentialCris <- which( crisRle$values == T )
  crisisDat <- list()
  while( currentInd <= nCris ){

    # Find the start and end of the crisis
    thisCris <- potentialCris[ currentInd ]
    startLoc <- sum( crisRle$lengths[1:(thisCris - 1)] ) + 1
    endLoc <- sum( crisRle$lengths[1:thisCris] )
  
    # skip crises of length 1
    if( endLoc >= startLoc + minLength-1 ){
      # check that the crisis end is correctly defined
      if( (endLoc < nData) && isCrisis[ endLoc + 1 ] )
        stop( "Crisis ending incorrectly defined" )
      # If there are fewer than minGap non-crisis periods between this
      # crisis and the next, merge the two crises.
      while( (endLoc < (nData-1)) && any( isCrisis[ (endLoc + 2):(endLoc + minGap) ] ) ){
         currentInd <- currentInd + 1
         thisCris <- potentialCris[ currentInd ]
         endLoc <- sum( crisRle$lengths[1:thisCris] )
      }
  
      # Store the data in crisisDat.
      # Include preCrisis time periods before the start of the
      # crisis and at most crisCutoff time periods after the start of the
      # crisis.
      endLoc <- min( startLoc + crisCutoff - 1, endLoc )
      startLoc <- max( 1, startLoc - preCrisis )
      thisDat <- metricsThresh[startLoc:endLoc,]
      crisisDat <- c( crisisDat,
                     list( thisDat ) )
    }
    currentInd <- currentInd + 1
  }
  crisisDat
}
