
#  Functions for posterior summaries


getMcmc <- function(
  outFile = "dp",  # The base name for the files containing the
                     # posterior samples.
  coda.loc = NULL   # the directory for the coda library
  ){

  #######################################################################
  # getMcmc
  #
  # Creates an mcmc object (used by the coda package) containing
  # the posterior samples for M (the number of atoms in the levy process)
  # and the outcome precision.
  #

  # Read output files
  outFileLG = paste(outFile,"Out.LG",sep="");   # Output Log (dput/dget)
  outFileMC = paste(outFile,"Out.MC",sep="");   # Output from MCMC
  if(!file.exists(outFileLG))
    stop(paste("Unable to read log file", outFileLG));
  if(!file.exists(outFileMC))
    stop(paste("Unable to read MCMC output file", outFileMC));
  MyInfo = dget(outFileLG);
  nCrises <- MyInfo$nCrises
  temp <- MyInfo$temp
  nTemps <- length(temp);

  fdat <- readBin(outFileMC, what="double", n=1e8, size=8);   # Fixed-length data
  fdat <- matrix(fdat, ncol = nCrises*nTemps, byrow = T);

  # load coda library if not yet loaded
  if( !is.element( "coda", search() ) ){
    if( is.null( coda.loc ) )
      library(coda);
    else {
      library(coda, lib.loc = coda.loc);
    }
  }

  # create an MCMC object 
  dpmc <- mcmc(fdat);  # as.matrix is only necessary here
                             # because of a bug in the mcmc function when run
                             # in Linux
  
  vars <- paste( "crisis:", (1:nCrises) );
  vars <- rep( vars, nTemps );
  vars <- paste( "Temp:", rep( temp, each = nCrises ), vars );
  varnames(dpmc) <- vars;
    
  dpmc
}


getAuxMcmc <- function(
  outFile = "dp",  # The base name for the files containing the
                     # posterior samples.
  coda.loc = NULL   # the directory for the coda library
  ){

  #######################################################################
  # getAuxMcmc
  #
  # Creates an mcmc object (used by the coda package) containing
  # the posterior samples of the auxiliary parameters
  #

  # Read output files
  outFileLG = paste(outFile,"Out.LG",sep="");   # Output Log (dput/dget)
  outFileAX = paste(outFile,"Out.AX",sep="");   
  if(!file.exists(outFileLG))
    stop(paste("Unable to read log file", outFileLG));
  if(!file.exists(outFileAX))
    stop(paste("Unable to read MCMC output file", outFileAX));
  MyInfo = dget(outFileLG);
  nCrises <- MyInfo$nCrises
  nMetr <- MyInfo$nMetrics

  alldat <- readBin( outFileAX, what="double", n=1e7, size=8);
  totLen <- length(alldat)
  fdat <- matrix(0, ncol = nCrises*nMetr*12+1, nrow = 0);
  loc <- 1
  
  while( loc < totLen ){
    nTypes <- alldat[loc]
    loc = loc+1
    thisSamp <- nTypes
    for( i in 1:nTypes ){
      # for each crisis type
      for( j in 1:nMetr ){
        # for each metric
        # First, get beta
        beta <- alldat[loc:(loc+2)];
        loc <- loc + 3
        # then get the transition matrix
        transMat <- alldat[loc:(loc+8)];
        loc <- loc + 9;
        thisSamp <- c( thisSamp, beta, transMat )
      }
    }
    # fill in thisSamp so that it has correct length
    thisSamp <- c( thisSamp, rep( 0, (nCrises-nTypes)*nMetr*12 ) )
    fdat <- rbind( fdat, thisSamp )
  }

  # load coda library if not yet loaded
  if( !is.element( "coda", search() ) ){
    if( is.null( coda.loc ) )
      library(coda);
    else {
      library(coda, lib.loc = coda.loc);
    }
  }

  # create an MCMC object
  fdat <- as.data.frame(fdat)
  dpmc <- mcmc(as.matrix(fdat));  # as.matrix is only necessary here
                             # because of a bug in the mcmc function when run
                             # in Linux
  
  dpmc
}

predictTypeOld <- function
(inFile = "dp",  # The base name for the files containing the
                     # posterior samples.
 newDat,           # the new data
 printInfo = T     # whether or not to print helpful info.
 ){

  #######################################################################
  # predictType
  #
  # Given the posterior distribution of the crisis types for previous
  # crises, and the data for a new crisis, calculates the approximate 
  # Pr( Z_new = Z_i | all data ) and Pr( Z_new != Z_i \forall i | all data )
  #

  # Read output files
  inFileLG = paste(inFile,"Out.LG",sep="");   # Output Log (dput/dget)
  inFileMC = paste(inFile,"Out.MC",sep="");   
  inFileAX = paste(inFile,"Out.AX",sep="");   
  if(!file.exists(inFileLG))
    stop(paste("Unable to read log file", inFileLG));
  if(!file.exists(inFileMC))
    stop(paste("Unable to read MCMC output file", inFileMC));
  if(!file.exists(inFileAX))
    stop(paste("Unable to read MCMC output file", inFileAX));
  MyInfo = dget(inFileLG);
  nCrises <- MyInfo$nCrises
  nMetr <- MyInfo$nMetrics
  nTemps <- length(MyInfo$temp)
  alpha <- MyInfo$alpha
  firstPrior <- MyInfo$firstPrior
  transPrior <- MyInfo$transPrior

  probVec <- rep(0, nCrises+1)
  newLen <- dim(newDat)[1]
  newDat <-  as.vector( as.matrix( newDat ) )

  res <- .C( "predictTypeOld", newDat = as.double( newDat ),
            newLen = newLen,
            mcFile = inFileMC, auxParamFile = inFileAX,
            nCrises = nCrises, nMetr = nMetr,
            nTemps = nTemps, alpha = alpha,
            firstPrior = as.double( firstPrior ),
            transPrior = as.double( as.vector( transPrior ) ),
            printInfo = printInfo,
            probVec = as.double(probVec) )

  return( res$probVec )
}

predictType <- function
(inFile = "dp",  # The base name for the files containing the
                     # posterior samples.
 newDat,           # the new data
 printInfo = T     # whether or not to print helpful info.
 ){

  #######################################################################
  # predictType
  #
  # Given the posterior distribution of the crisis types for previous
  # crises, and the data for a new crisis, calculates the approximate 
  # Pr( Z_new = Z_i | all data ) and Pr( Z_new != Z_i \forall i | all data )
  #

  # Read output files
  inFileLG = paste(inFile,"Out.LG",sep="");   # Output Log (dput/dget)
  inFileMC = paste(inFile,"Out.MC",sep="");   # crisis type post. samples file
  if(!file.exists(inFileLG))
    stop(paste("Unable to read log file", inFileLG));
  if(!file.exists(inFileMC))
    stop(paste("Unable to read MCMC output file", inFileMC));
  MyInfo = dget(inFileLG);
  nCrises <- MyInfo$nCrises
  nMetr <- MyInfo$nMetrics
  nTemps <- length(MyInfo$temp)
  alpha <- MyInfo$alpha
  firstPrior <- MyInfo$firstPrior
  transPrior <- MyInfo$transPrior

  probVec <- rep(0, nCrises+1)
  newLen <- dim(newDat)[1]
  newDat <-  as.vector( as.matrix( newDat ) )

  # Get the historical crisis data
  crisisDat <- MyInfo$Data
  crisLen <- rep( 0, nCrises )
  for( i in 1:nCrises )
    crisLen[i] <- dim( crisisDat[[i]] )[1]

  # Change crisisDat to a vector in order to pass to C code (cannot
  # pass it as a list of data frames)
  crisVec <- rep( -1, sum( crisLen * nMetr ) )
  currentLoc <- 1
  for( i in 1:nCrises ){
    for( j in 1:crisLen[i] ){
      for( k in 1:nMetr ){
        crisVec[currentLoc] <- crisisDat[[i]][j,k]
        currentLoc = currentLoc + 1
      }
    }
  }
  
  res <- .C( "predictType", newDat = as.double( newDat ),
            newLen = newLen, crisisDat = as.double(crisVec),
            crisLen = as.double(crisLen),
            mcFile = inFileMC, 
            nCrises = nCrises, nMetrics = nMetr,
            nTemps = nTemps, alpha = as.double(alpha),
            firstPrior = as.double( firstPrior ),
            transPrior = as.double( as.vector( transPrior ) ),
            printInfo = printInfo,
            probVec = as.double(probVec) )

  return( res$probVec )
}

nClusters <- function
( dpmc                       # The mcmc object containing the samples of the crisis types
 ){

  # first find the temperatures
  vars <- varnames(dpmc)
  temp <- c()
  for( i in 1:length(vars) ){
    w <- regexpr("[0-9]+.?[0-9]*", varnames(dpmc)[i] )
    thisTemp <- as.numeric(substring(varnames(dpmc)[i], w, w + attr(w, "match.length")-1 ))
    temp <- c(temp, thisTemp)
  }
  temp <- unique(temp)
  nTemps <- length(temp)
  nCrises <- nvar(dpmc) / nTemps

  # For the first temperature, calculate the number of clusters
  nSamps <- dim(dpmc)[1]
  nClust <- matrix(0, nrow = nSamps, ncol = nTemps)
  for( i in 1:nTemps ){
    for( j in 1:nSamps ){
      # find the column indices in dpmc that correspond to this temperature 
      thisInd <- (1:nCrises) + (i-1)*nCrises
      nClust[j,i] <- max( dpmc[j,thisInd] ) + 1  # Account for zero indexing
    }
  }
  nClust
}

pairFrequency <- function
( dpmc                       # The mcmc object containing the samples of the crisis types
 ){

  # first find the temperatures
  vars <- varnames(dpmc)
  temp <- c()
  for( i in 1:length(vars) ){
    w <- regexpr("[0-9]+.?[0-9]*", varnames(dpmc)[i] )
    thisTemp <- as.numeric(substring(varnames(dpmc)[i], w, w + attr(w, "match.length")-1 ))
    temp <- c(temp, thisTemp)
  }
  temp <- unique(temp)
  nTemps <- length(temp)
  nCrises <- nvar(dpmc) / nTemps

  # Calculate the frequency with which each pair of crises appear together for the first
  # temperature

  pairFreq <- matrix( nrow = nCrises, ncol = nCrises )
  for( i in 1:nCrises ){
    for( j in 1:nCrises ){
      crisI <- dpmc[,i]
      crisJ <- dpmc[,j]
      pairFreq[i,j] <- mean( crisI == crisJ )
    }
  }
  pairFreq
}

clusterMode <- function
( dpmc                       # The mcmc object containing the samples of the crisis types
 ){

  # first find the temperatures
  vars <- varnames(dpmc)
  temp <- c()
  for( i in 1:length(vars) ){
    w <- regexpr("[0-9]+.?[0-9]*", varnames(dpmc)[i] )
    thisTemp <- as.numeric(substring(varnames(dpmc)[i], w, w + attr(w, "match.length")-1 ))
    temp <- c(temp, thisTemp)
  }
  temp <- unique(temp)
  nTemps <- length(temp)
  nCrises <- nvar(dpmc) / nTemps
  nSamps <- dim( dpmc )[1]

  clustAssign <- unique( as.data.frame( as.matrix( dpmc[,1:nCrises] ) ) )
  clustAssign <- as.matrix( clustAssign )
  nAssign <- dim(clustAssign)[1]
  assignCounts <- rep(0, nAssign)
  for( i in 1:nAssign ){
    a <- matrix(data = as.vector(clustAssign[i,]), nrow=nSamps,
                ncol=nCrises, byrow = T)
    assignCounts[i] <- sum( rowSums( a == as.matrix( dpmc[,1:nCrises] ) ) == nCrises )
  }
  clustOrder <- rev( order( assignCounts ) )
  assignCounts <- assignCounts[clustOrder]
  clustAssign <- clustAssign[clustOrder,]
  return ( list( assignCounts = assignCounts, clustAssign = clustAssign ) )
  
}

#auxPost <- function
#( dpmc,
#  crisisDat
# ){
#
#  # calculate the transition counts for each crisis
#  crisisCounts <- list()
#  nCris <- length( crisisDat )
#  nMetr <- dim( crisisDat[[1]] )[2]
#  for( i in 1:nCris ){
#    # for each crisis
#    crisLen <- dim( crisisDat[[i]] )[1]
#    thisCris <- list()
#    for( j in 1:nMetr ){
#      # for each metric
#      thisCrisMetr <- matrix( 0, nrow=3, ncol=3 )
#      for( k in 1:(crisLen-1)){
#        # for each transition
#        thisVal <- crisisDat[[i]][k,j]
#        nextVal <- crisisDat[[i]][k+1,j]
#        thisCrisMetr[thisVal,nextVal] <- thisCrisMetr[thisVal,nextVal] + 1
#      }
#      thisCris <- c( thisCris, list( thisCrisMetr ) )
#    }
#    crisisCounts <- c( crisisCounts, list( thisCris ) )
#  }
#
#  
#  
#  return( crisisCounts )
#}
