

############################################################################
##
##  priorPostClust
##
## Validate the code for the clustering model given
## in Woodard and Goldszmidt (2010), by sampling from the prior,
## the data, then the posterior; this should be marginally distributed
## according to the prior.
##

priorPostClust <- function
(verbose = FALSE,  # whether to print details
 n.reps=20,       # the number of repetitions
 coda.loc = NULL   # the directory for the coda library
 ){

  # load coda library if not yet loaded
  if( !is.element( "coda", search() ) ){
    if( is.null( coda.loc ) )
      library(coda);
    else {
      library(coda, lib.loc = coda.loc);
    }
  }

  ## Create the inputs to the Cook and Rubin validation
	
  # Specify the hyperparameter values.  Theoretically any values for these 
  # will lead to a valid test.  

  alpha <- 5
  firstPrior <- c(.187,.632,.181)*2
  transPrior <- 5 * matrix( c(0.7, 0.2, 0.1, 0.08, 0.77, 0.15, 0.2, 0.2, 0.6),
         nrow = 3, ncol = 3, byrow = T )

  # the number of crises
  nCris <- 27
  nMetr <- 5
  crisLen <- 6

  # Markov chain parameters
  nIter <- 3e2
  nBurnin <- 1e2
  temps <- c(1, 1.7)
  
  hyper <- list( alpha = alpha, firstPrior = firstPrior, transPrior = transPrior )

  # the inputs for the parameter generation function
  generate.param.inputs <- list(hyper, nCris, verbose)
	
  # the inputs for the data generation function
  generate.data.inputs <- list( hyper, nMetr, crisLen, verbose)
	
  # the inputs for the function to draw from the posterior 
  analyze.data.inputs <- list(hyper, nIter, nBurnin, temps, verbose)

  ## run the validation
  #set.seed(11)
  nParams <- length( generate.param.DistClust( generate.param.inputs ) )
  priorSamp <- matrix( nrow = n.reps, ncol = nParams )
  postSamp <- matrix( nrow = n.reps, ncol = nParams )
  
  for( i in 1:n.reps ){

    priorSamp[i,] <- generate.param.DistClust( generate.param.inputs )
    dat <- generate.data.DistClust( priorSamp[i,], generate.data.inputs )
    postSamp[i,] <- analyze.data.DistClust( dat, priorSamp[i,],
                                      analyze.data.inputs )

  }

  pvals <- rep(0, nParams)
  for( j in 1:nParams ){
    pvals[j] <- ks.gof( x = priorSamp[,j], y = postSamp[,j] )$p.value
  }

  nParamsPlot = nParams
  maxPlot <- 10
  nParamsPlot <- min( nParamsPlot, maxPlot )
  par( mfrow = c( ceiling( nParamsPlot / 2 ), 2 ) )
  for( j in 1:nParamsPlot ){
    priorDens <- density( priorSamp[,j] )
    postDens <- density( postSamp[,j] )
    xlim <- range( c( priorDens$x, postDens$x ) )
    ylim <- range( c( priorDens$y, postDens$y ) )
    plot( priorDens$x, priorDens$y, type = "l", col = 1 )
    lines( postDens$x, postDens$y, type = "l", col = 2 )
  }
  if( verbose ){
    print( "prior samples: " ); print( priorSamp );
    print( "post. samples: " ); print( postSamp );
  }
  print( "p-values: " ); print( pvals );
  
  return( pvals )
}



# sample from the Dirichlet prior on the crisis parameters

# rdirichlet generates n random draws from the Dirichlet at
# vector x given shape parameter vector (or matrix)
# alpha.
#
# note: this code is taken verbatim from the R-package
# "Greg's Miscellaneous Functions" maintained by
# Gregory R. Warnes <Gregory_R_Warnes@groton.pfizer.com>
#
# Kevin Rompala 5/6/2003

"rdirichlet" <-
  function(n, alpha) {
    l <- length(alpha)
    x <- matrix(rgamma(l*n,alpha),ncol=l,byrow=TRUE)
    sm <- x%*%rep(1,l)
    return(x/as.vector(sm))
  }


#########################################################################
##  
##  generate.param.DistClust
##
##  generate the parameters for the model from a finite mixture model
##  


generate.param.finiteMix <- function(inputs){

  # get the hyperparameters
  nCris <- inputs[[2]]
  verbose <- inputs[[3]]
  
  # sample the parameters
  nClust <- rdiscrete( 1, rep( 1/nCris, nCris ) )
  clustProbs <- rdirichlet( 1, rep(1, nClust) )
  clustInd <- rdiscrete( nCris, clustProbs )
  clustInd <- orderTypeVec( clustInd )
  trueNclust <- length( unique( clustInd ) )
  
  if (verbose){
    # print the parameter values
    print("*******************************************************************")
    print("")
    print(paste("# clusters: ", trueNclust ) )
    print("cluster indicators: "); print( clustInd )
    print("")
  }

  params <- c( trueNclust, clustInd )
	
  return(params)
}


#########################################################################
##  
##  generate.param.DistClust
##
##  generate the parameters for the model from their prior distributions
##  
##  this function is used as input to the Cook and Rubin validation
##

generate.param.DistClust <- function(inputs){

  # get the hyperparameters
  hyper <- inputs[[1]]
  alpha <- hyper$alpha
  nCris <- inputs[[2]]
  verbose <- inputs[[3]]
  
  # sample the parameters
  clustInd <- samplePrior( nObs = nCris, alpha = alpha )
  nClust <- max(clustInd)
  
  if (verbose){
    # print the parameter values
    print("*******************************************************************")
    print("")
    print(paste("# clusters: ", nClust ) )
    print("cluster indicators: "); print( clustInd )
    print("")
  }

  params <- c( nClust, clustInd )
	
  return(params)
}


#########################################################################
##  
##  generate.data.DistClust
##
##  given the parameters of the DistClust, generate data from the model
##  
##  this function is used as input to the Cook and Rubin validation
##

generate.data.DistClust <- function(params, inputs){
	
  # get the prior counts 
  hyper <- inputs[[1]]
  firstPrior <- hyper$firstPrior
  transPrior <- hyper$transPrior
  # get the number of metrics
  nMetr <- inputs[[2]]
  # get the crisis length
  crisLen <- inputs[[3]]
  # get an indicator of whether to print information
  verbose <- inputs[[4]]
  # get the crisis indicators
  clustInd <- params[ 2:length(params) ]
  nCris <- length(clustInd)
  nTypes <- max( clustInd )

  # Sample the probabilities of the first values of each 
  # metric for each crisis type, and the probabilities
  # of the transitions
  firstProbs <- list()
  transProbs <- list()
  thisProbs <- matrix( ncol = 3, nrow = nMetr )
  thisMat <- array( dim = c(nMetr, 3, 3) )
  for( i in 1:nTypes ){
    for( j in 1:nMetr ){
      thisProbs[j,] <- rdirichlet( n=1, alpha=firstPrior )
    }
    firstProbs <- c( firstProbs, list(thisProbs) )
    for( j in 1:nMetr ){
      for( k in 1:3 ){
        thisMat[j,k,] <- rdirichlet( n=1, alpha=transPrior[k,] )
        if(abs(1 - sum(thisMat[j,k,])) > 0.001)
	  stop("the probabilities must sum to 1")
      }
    }
    transProbs <- c( transProbs, list(thisMat) )
  }

  if (verbose){
    # print these parameters
    print("*******************************************************************")
    print("")
    print("initial prob.s: "); print( firstProbs )
    print("")
    print("transition prob.s: "); print( transProbs )
    print("")
  }

  # Sample the crisis data
  crisDat <- list()
  thisDat <- matrix( -1, nrow = crisLen, ncol = nMetr )

  for( i in 1:nCris ){
    thisProbs <- firstProbs[[ clustInd[i] ]]
    thisMat <- transProbs[[ clustInd[i] ]]
    for( j in 1:nMetr ){
#      if( any( thisProbs[j,] > 0.999 ) )
#        samp <- which( thisProbs[j,] > 0.999 )
#      else
      samp <- rdiscrete(n=1, prob = thisProbs[j,] )
      thisDat[1,j] <- samp - 1   # need zero indexing for call to MCMC routine
      if( crisLen > 1 ){
        for( k in 1:(crisLen-1) ){
          currVal <- thisDat[k,j] + 1
          samp <- rdiscrete(n=1, prob=thisMat[j,currVal,])
          if( samp == 0 ){
            print("thisMat[j,currVal,] = ");  print(thisMat[j,currVal,]);
            stop("0 value sampled");
          }
          thisDat[k+1,j] <- samp - 1
        }
      }
    }
    crisDat <- c( crisDat, list( thisDat ) )
  }
	
  if (verbose){
    # print the data
    print("*******************************************************************")
    print("")
    print("data: "); print( crisDat )
    print("")
  }

  return( crisDat )
}


#########################################################################
##  
##  analyze.data.DistClust
##
##  Obtain posterior samples from the model.
##  
##  this function is used as input to the Cook and Rubin validation
##

analyze.data.DistClust <- function(y, params.true, inputs){
		
  # get the hyperparameters
  hyper <- inputs[[1]]
  # get the # of iterations
  nIter <- inputs[[2]]
  # get the burn-in length
  nBurnin <- inputs[[3]]
  # get the temperatures
  temps <- inputs[[4]]
  # get an indicator of whether to print information
  verbose <- inputs[[5]] 

  # get the posterior samples
  post.mcmc <- sampleDistClust (data = y, nIter = nIter,
                                nBurnin = nBurnin, temps = temps,
                                hyper = hyper)

  if (verbose)	{
    if (nvar(post.mcmc) > 30){
      tmp.post.mcmc <- post.mcmc[,1:30]
    } else {
      tmp.post.mcmc <- post.mcmc
    }
    print(summary(tmp.post.mcmc))
  }
  # check the convergence of the chain.  If not converged, print
  # trace plots and issue a warning
  #poorConvergeParams <- (effectiveSize(post.mcmc) < niter(post.mcmc)/10)
  #if (any(poorConvergeParams)){
  #  warning("Poor convergence of the some parameters.  Traceplot generated.")
  #  print(effectiveSize(post.mcmc))
  #  par(mfrow = c(3,2))
  #  print(traceplot(post.mcmc[,poorConvergeParams]))
  #}

  # return the last posterior sample
  return ( post.mcmc[ niter(post.mcmc), ] )
}

sampleDistClust <- function(data, nIter, nBurnin, temps, hyper){

  # specify the initial crisis type vector
  #typeVec = (1:length(data))-1
  typeVec = rep(0, length(data))

  alpha <- hyper$alpha
  firstPrior <- hyper$firstPrior
  transPrior <- hyper$transPrior

  dpMcmc(data, typeVec,
         alpha = alpha, firstPrior = firstPrior,
         transPrior = transPrior,
         nIter = nIter, nBurnin= nBurnin,
         outFile = "temp", temp = temps, printStats = F)

  crisTypes <- getMcmc(outFile = "temp")
  nCrises <- nvar( crisTypes ) / 2
  nClust <- as.vector( nClusters(crisTypes) )
  crisTypes <- crisTypes + 1     # Correct for zero indexing of crisis labels from MCMC
  
  return( mcmc( cbind( nClust, crisTypes[,1:nCrises] ) ) )
}



