
simulateTS <- function(
  # the probability of moving into a crisis period,
  # given that you are in a non-crisis period:
  crisTransProb = 0.003,
  # The parameters of the neg. binomial prior for the
  # length of a crisis period:
  crisLenShape = 3000,
  crisLenRate = 100,
  # The prob.s of each type of crisis:
  crisProbs = c(0.2, 0.45, 0.35),
  # The number of metrics:
  nMetrics = 8,
  # The number of Key Performance Indicators:
  nKPIs = 2,
  # The rate parameter for the Poisson prior on the
  # number of metrics affected by crises of a particular
  # type:
  nAffectRate = 0.8,
  # The rate parameter for the Poisson prior on the
  # number of events per time period in the affected metrics during crisis
  # periods:
  nEventRate = 0.4,
  # The shape and rate parameters for the kernel magnitudes:
  kernMagShape = 60,
  kernMagRate = 100,
  # The priors for the AR-1 processes of the
  # metrics during background periods. These AR-1 proc.s
  # are defined by a mean, a lag-1 autocorrelation, and
  # an incremental variance.  These will be given
  # normal, beta, and inverse gamma prior distributions
  # respectively.  Define the hyperparameters of these
  # prior distributions:
  meanMu = -2,
  meanTau2 = 0.05,
  autocorrShape1 = 100,
  autocorrShape2 = 20,
  arVarShape = 300,
  arVarRate = 10,
  # The kernel scale:
  kernScale = 6,
  # The time series length
  tsLength = 4000 
){


  # function to evaluate the density of a kernel with center given
  # by "center" and scale given by "scale"
  kernelEval <- function(loc, center, scale = 2){
    kernDens <- (loc - center)^2
    kernDens <- -1 * kernDens / (2 * scale^2)
    kernDens <- exp(kernDens)
    return(kernDens)
  }

  ################### Sample the Crises #########################################
  
  # Sample the metrics affected by each
  # type of crisis, and the direction and magnitude of the effect 
  numTypes <- length( crisProbs )
  nCrisAff <- rep( 0, numTypes )
  nMetrKPIs <- (nKPIs + nMetrics)
  crisAffect <- matrix( 0, nrow = numTypes, ncol = nMetrKPIs )
  for( j in 1:numTypes ){
    # sample the number of KPIs affected by this type of
    # crisis
    nCrisAff[j] <- rpois( n=1, lambda = nAffectRate ) + 1
    nCrisAff[j] <- min( nCrisAff[j], nKPIs )
    crisAffInds <- sample( n = nKPIs, size = nCrisAff[j] )
    for( k in 1:nCrisAff[j] ){
        crisAffect[ j, crisAffInds[k] ] <-
          rgamma(n=1, shape=kernMagShape, rate=kernMagRate)
    }
    # sample the number of metrics affected by this type of
    # crisis
    nCrisAff[j] <- rpois( n=1, lambda = nAffectRate ) + 1
    nCrisAff[j] <- min( nCrisAff[j], nMetrics )
    crisAffInds <- sample( n = nMetrics, size = nCrisAff[j] )
    for( k in 1:nCrisAff[j] ){
      if( runif(n=1) < 0.5 )
        crisAffect[ j, (nKPIs + crisAffInds[k]) ] <-
          rgamma(n=1, shape=kernMagShape, rate=kernMagRate)
      else
        crisAffect[ j, (nKPIs + crisAffInds[k]) ] <- -1 *
          rgamma(n=1, shape=kernMagShape, rate=kernMagRate)
    }
  }
  
  # sample the parameters of the background AR-1 metric dist'ns;
  # these each have a mean, a correlation rho, and an incremental
  # variance sigma^2
  metrMeans <- rnorm( n=nMetrKPIs, mean = meanMu, sd = sqrt(meanTau2) )
  metrCorrs <- rbeta( n=nMetrKPIs, shape1=autocorrShape1,
                     shape2=autocorrShape2 )
  metrSigma2s <- 1/ rgamma( n=nMetrKPIs, shape = arVarShape,
                           rate = arVarRate )


  # sample the crisis / non-crisis indicators for each time
  # period, the crisis types, and the crisis kernel locations.
  inCrisis <- rep(0, tsLength)
  crisTimeLeft <- 0
  numCris <- 0
  crisType <- c()
  isKernelLoc <- rep(0, tsLength)
  for( i in 2:tsLength ){
    if( crisTimeLeft > 0 )
      inCrisis[i] = 1
    if( ( inCrisis[i-1] == 0 ) && ( runif(n=1) < crisTransProb ) ){
      inCrisis[i] <- 1
      crisTimeLeft <- rnegbin( n=1, shape = crisLenShape,
                                rate = crisLenRate )
      numCris <- numCris + 1
      crisType <- c(crisType,
                    which( rdiscrete( n=1, prob = crisProbs ) ) )
      nKerns <- rpois( n=1, lambda= crisTimeLeft*nEventRate )
      kernInds <- sample( x = i:(i+crisTimeLeft-1), size = nKerns )
      # remove the kernel indicators that are outside the domain
      # of the time series
      inTS <- (kernInds <= tsLength)
      kernInds <- kernInds[inTS]
      isKernelLoc[ kernInds ] <- 1
    }
    if( inCrisis[i] )
      crisTimeLeft <- crisTimeLeft - 1
  }
  # check that the kernels are only located in crisis periods
  if( !as.logical( sum((inCrisis - isKernelLoc) < 0) == 0 ) )
    stop("Kernel locations not all during crisis periods")

  # simulate the metric time series.  These are independent.
  tsMean <- matrix( nrow = nMetrKPIs, ncol = tsLength )
  for( k in 1:nMetrKPIs ){
    for( i in 1:tsLength ){
      tsMean[k,i] <- metrMeans[k]
  
    }
  }
  
  crisInd <- 0
  for( i in 2:tsLength ){
  
    if( inCrisis[i] && !inCrisis[i-1] ){
      crisInd <- crisInd + 1
      thisCrisType <- crisType[crisInd]
    }
    
    if( isKernelLoc[i] ){
      startLoc <- max(1, round(i-3*kernScale) )
      endLoc <- min(tsLength, round(i+3*kernScale) )
      for( j in startLoc:endLoc ){
        for( k in 1:nMetrKPIs )
          tsMean[k,j] <- tsMean[k,j] +
            crisAffect[ thisCrisType, k ] *
              kernelEval(j, center = i, scale = kernScale)
      }
    }
  }


  metricsTS <- matrix( nrow = nMetrKPIs, ncol = tsLength )
  for( k in 1:nMetrKPIs ){
    metricsTS[k,1] <- metrMeans[k]
  }
  for( i in 2:tsLength ){
    for( k in 1:nMetrKPIs ){
      incr <- rnorm( n=1, sd=sqrt(metrSigma2s[k]) )
      metricsTS[k,i] <- metricsTS[k,i-1] - tsMean[k,i]
      metricsTS[k,i] <- metrCorrs[k] * metricsTS[k,i]
      metricsTS[k,i] <- metricsTS[k,i] + tsMean[k,i]
      metricsTS[k,i] <- metricsTS[k,i] + incr
    }
  }
  
  # transform to non-negative scale
  for( i in 1:tsLength ){
    for( k in 1:nMetrKPIs ){
      metricsTS[k,i] <- exp( metricsTS[k,i] )*300
      metricsTS[k,i] <- round(metricsTS[k,i])
    }
  }

  rownames(metricsTS) <- as.character(1:nMetrKPIs)
  for( k in 1:nKPIs ){
    rownames(metricsTS)[k] <- paste( "KPI", k )
  }
  for( k in 1:nMetrics ){
    rownames(metricsTS)[nKPIs + k] <- paste( "Metric", k )
  }    

  dat <- as.data.frame( t( metricsTS ) )
  
  return( list(dat = dat, crisTypes = crisType) )
}



rnegbin <- function( n = 1, shape = stop("No shape argument"), rate = 1 ){
# Sample from a negative binomial
# distribution with a particular shape and rate

  poisRate <- rgamma( n=n, shape = shape, rate = rate )
  return( rpois( n = n, lambda = poisRate ) )
}

