\name{dpMcmc}
\alias{dpMcmc}
\title{Run the Markov chain}
\description{Run the
  Markov chain to sample from the 
  posterior distribution of the time series clustering model, as
  described in 
  Woodard and Goldszmidt (2010).  }
\details{
There are also help files in this package for the functions: getMcmc,
nClusters, pairFrequency, priorPostClust, and simulateData.
}
\seealso{\link{getMcmc},\link{nClusters},\link{pairFrequency},\link{priorPostClust},\link{simulateTS}
}
\usage{
dpMcmc( tsDat, clustInit, 
 firstPrior = rep(1, 3), 
 transPrior = matrix(1, nrow=3, ncol=3),  
 alpha = 1.0, temp = 1.0, 
 nIter = 10000, nBurnin = 1000,
 nSave = min( nIter, 1000 ), outFile = "dp",
 printStats = T)
}
\arguments{
  \item{tsDat}{The time series observations, formatted as a list of matrices
    containing the time series data for each observation.  For the
    example of crisis identification in distributed computing, this
    object can be generated by
    a call to \code{crisisList}.}
  \item{clustInit}{A vector of initial cluster assignments (integers), of length equal to
    the number of observations in \code{tsDat}.  The cluster assignments
    can be indexed from 0 or from 1.}
\item{firstPrior}{A length-3 vector of hyperparameters for the Dirichlet prior
  distribution for the first value of a variable for time series in a particular
  cluster.  Called \bold{a^(j)} in Woodard and Goldszmidt (2010).}
\item{transPrior}{A 3x3 matrix of hyperparameters for the product Dirichlet prior
  distribution for the transition matrix of a variable for time series in a
  particular cluster.  Called \bold{b^(j)} in Woodard and Goldszmidt (2010).}
\item{alpha}{The parameter of the Dirichlet process that controls the
  expected number of clusters in a set of I observations.}
\item{temp}{ A vector of the temperatures to be used in the parallel
  tempering algorithm.  The default is no tempering, given by \code{temp
    = 1.0}.}
\item{nIter}{The number of iterations of the Markov chain, after burn-in
  and without accounting for thinning.}
\item{nBurnin}{The number of burn-in iterations of the Markov chain.}
\item{nSave}{The approximate number of samples to store.  The Markov chain will be
  thinned to achieve approximately this number of post-thinning
  samples.}
\item{outFile}{The directory and base file name for files storing the
  samples and logged information.  Two files are generated; that with
  extension ".MC" is a binary file containing the posterior samples of
  the cluster assignments, and that with
extension ".LG" is a text log file containing information about the
arguments to \code{dpMcmc}, such as the Markov chain control
parameters.}
\item{printStats}{Whether to print summary statistics regarding the
  Markov chain, such as the run time and number of proposals and acceptances of each
  move type.}
}
\examples{

# Here we cluster observed "crises" in a distributed computing center,
# using simulated data.  The cutoffs defining crises are:
kpi1Cutoff <- 80
kpi2Cutoff <- 120

# Next we obtain the time series observations in the correct format to
# pass to the dpMcmc function:
crisisDat <- crisisList( simulMetrics,
  1:2, c(kpi1Cutoff, kpi2Cutoff ))

# Obtain some initial cluster assignments by sampling from the DP prior
typeVec = samplePrior( alpha = 9, nObs = length(crisisDat) )

# The following is an illustrative run.  To obtain rigorous results,
# increase nIter and nBurnin, and choose firstPrior and transPrior
# as described in Woodard and Goldszmidt (2010).
transPrior <- matrix(0.1, ncol = 3, nrow = 3)
transPrior[1,1] <- 0.5;  transPrior[2,2] <- 0.5;  transPrior[3,3] <- 0.5;  
dpMcmc(crisisDat, typeVec,
       alpha = 9.0, firstPrior = rep(0.1, 3),
       transPrior = transPrior,
       nIter = 1e3, nBurnin= 1e2, nSave = 1e3,
       outFile = "test", temp = c(1, 1.7, 2.5) )

\dontrun{
# get the posterior samples of the cluster assignment vector.  To run
# the following line you will need to change the "coda.loc" directory.
crisTypes <- getMcmc(outFile = "test", coda.loc = "/home/fs01/dbw59/rlibs")


# check convergence
ess <- effectiveSize(crisTypes)
ess

# Obtain the number of clusters for each posterior sample,
# for the first temperature (1.0)
nClust <- nClusters(crisTypes)
hist(nClust)

# Obtain estimates of the cluster labels
summary(crisTypes)

# Create a plot showing the frequency with which each pair of
# crises have identical labels.
pf <- pairFrequency( crisTypes )
image( pf )
nCrises <- dim(pf)[1]
image.legend (pf, x=c(1,5), y=c(nCrises-3, nCrises-1))
}
}
