
#include "ClustEM.h"

/* debugging info: 
   To print high-level debugging info at runtime, add "-DDEBUG1" to the first line of the Makevars
   file. 
*/



#define SQR(x) ((x)*(x))

using TNT::Array1D;
using TNT::Array2D;
using std::cout;
using std::endl;

extern "C"
{
  double S_STDCALL S_rpois( double lambda );
  double S_STDCALL S_rgamma( double s );
  double S_STDCALL S_runif( void );
  double S_STDCALL S_rexp( void );
}


ClustEM::ClustEM
(Array2D<double> *crisisList,
 int nCris,
 const Array1D<double> &firstPrior,
 const Array2D<double> &transPrior
 )
{
  crisisDat_ = crisisList;
  nCris_ = nCris;
  nMetr_ = crisisList[0].dim2();

  modelFit = false;   // whether the cluster model has been fit.
  bestNclust_ = 0;
  bestClust_ = Array2D<double>();

  #ifdef DEBUG1
    cout << "Crisis list [0]" << crisisList[0] << endl;
    cout << "nMetr: " << nMetr_ << endl;
  #endif
  firstPrior_ = firstPrior.copy();
  transPrior_ = transPrior.copy();
  if( sum( firstPrior_ ) > 0 )
    usingPrior = true;
  else
    usingPrior = false;

  // Create the counts objects.  
  // Calculate the transition counts for each 
  // crisis and metric.
  firstCounts = new Array1D<double> * [ nCris_ ];
  transCounts = new Array2D<double> * [ nCris_ ];
  crisisCounts = new Array2D<double> * [ nCris_ ];

  int i,j, k1, crisLen, thisObs, nextObs;
  for( i=0; i<nCris_; i++ ){
    // For each crisis
    // create the object vectors
    firstCounts[i] = new Array1D<double> [ nMetr_ ];
    transCounts[i] = new Array2D<double> [ nMetr_ ];
    crisisCounts[i] = new Array2D<double> [ nMetr_ ];
    
    // get the # of time periods in the crisis
    crisLen = crisisDat_[i].dim1();
    for( j=0; j<nMetr_; j++ ){
      // For each metric
      // create the objects
      // Don't have to create firstCounts[i][j]; this is handled
      // before use of transCounts during likelihood calculation.
      transCounts[i][j] = Array2D<double>( 3, 3, 0.0 );
      crisisCounts[i][j] = Array2D<double>( 3, 3, 0.0 );

      // Calculate the transition counts for this crisis and metric
      for( k1=0; k1<(crisLen-1); k1++ ){
	// for each time period in the crisis except the last	
	// find the transition from time k1 to time k1+1
	thisObs = (int) crisisDat_[i][k1][j];
	nextObs = (int) crisisDat_[i][k1+1][j];
	// increment the appropriate count
        crisisCounts [i][j][thisObs][nextObs] = crisisCounts [i][j][thisObs][nextObs] + 1;
      }
      #ifdef DEBUG1
        cout << "Crisis counts [" << i << "][" << j << "]" << crisisCounts[i][j] << endl;
      #endif
    }
  }
}

ClustEM::~ClustEM(){}



/**************************************************************************
 ** Order a vector of crisis type indicators (leave no gaps in indexing)
 **/  
Array1D<double> ClustEM::orderTypeVec( const Array1D<double> &crisType ){
  
  Array1D<double> crisLabels = unique( crisType );
  int i;
  Array1D<double> orderedTypes( nCris_, 0.0 );
  for( i=0; i<nCris_; i++ ){
    orderedTypes[i] = find( crisLabels, crisType[i] );
  }	
  return( orderedTypes );
}


/**************************************************************************
 ** Calculate the counts for the crisis types and store in firstCounts and 
 ** transCounts
 **/  
void ClustEM::calcCounts
( const Array1D<double> &crisType
){

  int nClust = unique( crisType ).dim();

  // Create a 2-D array of the probabilistic assignments
  Array2D<double> probAssign( nCris_, nClust, 0.0 );
  int crisInd;
  for( int i=0; i<nCris_; i++ ){
    crisInd = (int)crisType[i];
    probAssign[i][crisInd] = 1.0;
  }

  return( calcCounts( probAssign ) );
}


/**************************************************************************
 ** Calculate the counts for the crisis types, where the cluster
 ** assignments are probabilistic; store the resulting counts in
 ** firstCounts and transCounts.
 **/  
void ClustEM::calcCounts
( const Array2D<double> &clustAssign
){

  int i, j, k, numTypes = clustAssign.dim2();

  // Initialize the counts at their priors
  for( i=0; i<numTypes; i++ ){
    // For each crisis type

    for( j=0; j<nMetr_; j++ ){
      // For each metric:
      if( usingPrior ){
	firstCounts[i][j] = firstPrior_.copy();
	transCounts[i][j] = transPrior_.copy();
      } else {
	firstCounts[i][j] = Array1D<double> (3, 0.0);
	transCounts[i][j] = Array2D<double> (3, 3, 0.0);
      }
    }
  }
  
  // Calculate the counts

  int thisObs;
  for( i=0; i<nCris_; i++ ){
    // for each observation

    for( j=0; j<nMetr_; j++ ){
      // for each metric

      // find the first value of this metric during this crisis
      thisObs = (int) crisisDat_[i][0][j];

      for( k=0; k<numTypes; k++ ){
	// for each crisis type

	// increment the appropriate first count
	firstCounts [k][j][thisObs] += clustAssign[i][k];
      
	// add the transition counts for this crisis to the transition counts for this 
	// crisis type
	transCounts [k][j] = transCounts [k][j] + clustAssign[i][k] * crisisCounts[i][j];
      }
    }
  }
  #ifdef DEBUG_COUNTS
    cout << "crisis assignments: " << clustAssign << endl; 
    for( i=0; i<numTypes; i++ ){
      // For each crisis type

      for( j=0; j<nMetr_; j++ ){
        // For each metric:
        cout << " firstCounts i=" << i << " j=" << j << endl; 
	cout << firstCounts[i][j] << endl;
        cout << " transCounts: " << endl; 
        cout << transCounts[i][j] << endl;
      }
    }
  #endif 

}


/**************************************************************************
 **  given a cluster assignment, find the maximum classification
 **  log-likelihood.  clustAssign is the cluster assignment vector.
 **/
double ClustEM::maxLogLike
( const Array1D<double> &clustAssign )
{

  calcCounts( clustAssign );

  int i,j, nClust = unique(clustAssign).dim();

  // Initialize objects for the parameter estimates of a single cluster
  Array1D<double> mleGam[nMetr_];
  Array2D<double> mleT[nMetr_];
  for( j=0; j<nMetr_; j++ ){
    mleGam[j] = Array1D<double>( 3, 0.0 );
    mleT[j] = Array2D<double>( 3, 3, 0.0 );
  }

  double logLike = 0;
  for( j=0; j<nClust; j++ ){
    #ifdef DEBUG_COUNTS
      cout << "calculating max log likelihood for cluster " << j << endl;
    #endif
    // for each cluster
    // Find the MLEs for the cluster parameters (M-step)
    mapParams( j, mleGam, mleT );

    // update the log likelihood
    for( i=0; i<nCris_; i++ ){
      // for each observation
      // If the obs. belongs to this cluster, update the 
      // log-likelihood
      if( (int)clustAssign[i] == j )
	logLike += evalLike( i, mleGam,
			     mleT, true );
    }
  }
  #ifdef DEBUG_COUNTS
    cout << "Overall max log likelihood: " << logLike << endl;
  #endif

  //  delete [] mleGam;
  //delete [] mleT;
  return(logLike);
}


/**************************************************************************
 ** calculate the log likelihood ratio for merging clusters clust1 and
 ** clust2
 **/  
double ClustEM::likeRatio
( const Array1D<double> &clustAssign, 
  int clust1, int clust2 ){

  int i;

  // Create a merged cluster assignment vector
  Array1D<double> mergedClusts = clustAssign.copy();
  for(i=0; i<nCris_; i++){
    if( (int)mergedClusts[i] == clust2 )
      mergedClusts[i] = (double) clust1;
  }
  mergedClusts = orderTypeVec( mergedClusts );

  #ifdef DEBUG_COUNTS
    cout << "Current clusters" << clustAssign << endl;
    cout << "Merged clusters" << mergedClusts << endl;
  #endif
   
  double res = maxLogLike( mergedClusts ) - maxLogLike( clustAssign );

  return( res );
}


/**************************************************************************
 **  Apply model-based hierarchical clustering, which yields a
 **  hierarchy of cluster assignments; these are the rows of the
 **  returned matrix.
 **/
Array1D<double> *ClustEM::hierarClust(){

  // initially assign each observation to its own cluster
  Array1D<double> clustAssign( nCris_ );

  int i, j;
  for(i=0; i<nCris_; i++)
    clustAssign[i] = (double)i;

  // save the sequence of cluster assignments
  Array1D<double> *histClusts = new Array1D<double> [nCris_];
  for(i=0; i<nCris_; i++){
    histClusts[i] = Array1D<double> (nCris_, 0.0);
    histClusts[0][i] = clustAssign[i];
  }

  double lr, maxLR;
  double clust1=-1, clust2 =-1;
  int nClust = nCris_;
  // until there is a single cluster
  while( nClust > 1 ){

    //    #ifdef DEBUG1
      cout << "Current cluster assignment vector: " << clustAssign << endl;
      //#endif

    // find the merge that gives the smallest decrease in the
    // likelihood.  Save the cluster labels in bestPair.
    maxLR = -DBL_MAX;
    for( i=0; i< nClust; i++ ){
      for( j=(i+1); j<nClust; j++ ){
	// for each pair of existing clusters
	// calculate the change in the classification likelihood
	// associated with merging the clusters
	lr = likeRatio( clustAssign, i, j );
	if( lr > maxLR ){
	  clust1 = (double) i;
	  clust2 = (double) j;
	  maxLR = lr;
	}
      }
    }
    
    // merge the clusters
    for(i=0; i<nCris_; i++)
      if( clustAssign[i] == clust2 )
	clustAssign[i] = clust1;

    clustAssign = orderTypeVec(clustAssign);

    nClust--;
    
    // store the new cluster assignment
    for(i=0; i<nCris_; i++)
      histClusts[nCris_ - nClust][i] = clustAssign[i];

  }

  // return the cluster assignments
  return (histClusts);

}

/**************************************************************************
 **  Find cluster parameters that maximize the posterior density, for
 **  a particular cluster.  datClustPr is the current matrix of
 **  probabilistic assignments of observations to clusters, and
 **  clustInd is the index of the cluster of interest.  The MAP
 **  estimates are returned in mapGam and mapT.
 **/
void ClustEM::mapParams
( int clustInd,
  Array1D<double> *mapGam, 
  Array2D<double> *mapT ){

  int j,s,t;

  Array1D<double> this1stCounts;
  Array2D<double> thisTrans;

  double thisRowSum;
  for( j=0; j<nMetr_; j++ ){
    // For each metric

    // normalize the counts to get the MAP for gamma
    this1stCounts = firstCounts[clustInd][j];
    mapGam[j] =  this1stCounts / sum( this1stCounts );

    // normalize the transition counts to get the MAP for
    // each row of T
    thisTrans = transCounts[clustInd][j];

    for( s=0; s<3; s++ ){
      // for each row of the transition matrix

      // calculate the row sum
      thisRowSum = 0;
      for( t=0; t<3; t++ )
	thisRowSum += thisTrans[s][t];

      // normalize the row by its sum
      for( t=0; t<3; t++ ){
	if( thisRowSum > 0 )        
	  mapT[j][s][t] = thisTrans[s][t] / thisRowSum;
	else
	  mapT[j][s][t] = 0;
      }
    }

    #ifdef DEBUG_COUNTS
      cout << "gamma MAP for metric " << j << ": "<< mapGam[j] << endl;
      cout << "T MAP for metric " << j << ": " <<mapT[j] << endl;
    #endif

  }

}

/**************************************************************************
 **  Evaluate the likelihood for a single observation, given a set of
 **  parameter values.  obsInd is the index of the observation. 
 **/
double ClustEM::evalLike
  ( int obsInd,
    Array1D<double> *gam,
    Array2D<double> *T,
    bool logCalc ){

  double logLike = 0;
  int j, s, t, thisY;

  #ifdef DEBUG_COUNTS
    cout << "Evaluating the likelihood for obs. " << obsInd << endl;
  #endif

  for( j=0; j<nMetr_; j++ ){
    // for each metric

    // calculate the contribution of the first
    // value of the metric, Y_{i1j}
    // to the log-likelihood
    thisY = (int) crisisDat_[obsInd][0][j];
    logLike += log( gam[j][thisY] );

    // For each possible transition, calculate 
    // the contribution of this transition to the 
    // log-likelihood
    for( s=0; s<3; s++ )
      for( t=0; t<3; t++ ){
	if( crisisCounts[obsInd][j][s][t] > 0 )
	  logLike += crisisCounts[obsInd][j][s][t] * log( T[j][s][t] );
      }
    #ifdef DEBUG_COUNTS
      cout << "For metric " << j << " gamma = " << gam[j] << endl;
      cout << "T = " << T[j] << endl;
      cout << "Counts = " << crisisCounts[obsInd][j] << endl;
    #endif

  }
  if( logCalc )
    return( logLike );
  else	
    return( exp( logLike ) );
}

/**************************************************************************
 **    EM for clustering.  Requires an initial cluster assignment,
 **    given in the vector clusts.  Iterates until the improvement in
 **    log-likelihood is less than .01, i.e. until the likelihood
 **    changes by less than a factor of 1%.  Returns the
 **    log-likelihood in logLike.  The input vector clusts should have
 **    correctly ordered cluster indices.
 **/  
Array2D<double> ClustEM::emClust
( const Array1D<double> &clusts, 
  double &logLike )
{

  // find the # of clusters
  int i,j,k, s,t, clustInd,
    nClust = unique(clusts).dim();

  #ifdef DEBUG1
    cout << "EM: nClust = " << nClust << " and clusts vector = " << clusts << endl;
  #endif

  // initialize the matrices of probabilistic assigments
  // to clusters.  Leave a little probability in each category.
  Array2D<double> datClustPr(nCris_, nClust, 0.2/(nClust-1) ),
    datClustPrNew( nCris_, nClust, 0.0);
  for( i=0; i<nCris_; i++ ){
    clustInd = (int)clusts[i];
    if( nClust == 1 )
      datClustPr[i][clustInd] = 1.0;
    else
      datClustPr[i][clustInd] = 0.8;
  }

  // Declare the vector of cluster probabilities
  Array1D<double> clustProb;

  // Initialize objects for the parameter estimates of a single cluster
  Array1D<double> mapGam[nMetr_];
  Array2D<double> mapT[nMetr_];
  for( j=0; j<nMetr_; j++ ){
    mapGam[j] = Array1D<double>( 3, 0.0 );
    mapT[j] = Array2D<double>( 3, 3, 0.0 );
  }

  double oldLogPost = -DBL_MAX,
    logPostImp = DBL_MAX,
    newLogPost, thisLike;

  while( logPostImp > .01 ){
    // while the posterior density of the parameters ({p_j}_{j=1}^K, \gamma, T) 
    // continues to improve
    cout << "Probabilistic cluster assignment: " << datClustPr << endl;

    // Find the probabilities of the clusters (M-step)
    clustProb = TNT::colSums( datClustPr );

    // If we are using priors for the parameters, also add a count
    // of one to each cluster count
    if( usingPrior )
      for( k=0; k<nClust; k++ )
	clustProb[k] += 1.0;

    // normalize the cluster prob. vector
    clustProb = clustProb / sum(clustProb);

    #ifdef DEBUG_COUNTS
      cout << "Cluster probabilities: " << clustProb << endl;
    #endif

    // Find the MAPs of the cluster parameters.  
    // First, calculate the counts for each cluster.
    calcCounts(datClustPr);

    newLogPost = 0;
    for( k=0; k<nClust; k++ ){
      // for each cluster
      // Find the MAPs for the cluster parameters, using the
      // probabilistic cluster assignments (M-step).  Save
      // the prior log density of the parameters in newLogPost
      mapParams( k, mapGam, mapT );

      // update the log prior 
      if( usingPrior ){
	for( j=0; j<nMetr_; j++ )
	  for( t=0; t<3; t++ ){
	    newLogPost += (firstPrior_[t] - 1) * log( mapGam[j][t] );
	      for( s=0; s<3; s++ ){
		newLogPost += (transPrior_[s][t] - 1) * log( mapT[j][s][t] );
	      }
	  }
      }

      // Do the calculations to get the new probabilistic
      // cluster assignments (E-step)
      for( i=0; i<nCris_; i++ ){
	// for each observation
	// Calculate the likelihood of this observation for this 
	// cluster.
        datClustPrNew[i][k] = clustProb[k] *
          evalLike( i, mapGam,
		    mapT, false );
      }
    }

    // calculate the new posterior density, which is a product of the 
    // priors on ({p_j}_{j=1}^K, \gamma, T) and the mixture likelihood 
    // for each observation.  Also normalize the
    // probabilistic cluster assignments, which are assigned to
    // datClustPr
    for( i=0; i<nCris_; i++ ){
      // for each observation
      // update the log likelihood

      // The likelihood for this observation is the ith rowsum
      // of datClustPrNew.
      thisLike = 0;
      for( j=0; j<nClust; j++ ){
	thisLike += datClustPrNew[i][j];
      }
      newLogPost += log( thisLike ); 
      
      // normalize the cluster probabilities
      for( j=0; j<nClust; j++ ){
	datClustPr[i][j] = datClustPrNew[i][j] / thisLike;
      }
    }
    // find the change in log-likelihood
    logPostImp = newLogPost - oldLogPost;
    cout << "Log posterior density improvement in one EM iteration: " << logPostImp << endl;

    // save the log-likelihood
    oldLogPost = newLogPost;
  }

  cout << "Final cluster assignment: " << datClustPr << endl;

  // return the maximized mixture log-likelihood and the probabilistic
  // cluster assignment
  logLike = newLogPost;
  return (datClustPr);
}

/**************************************************************************
 **  Calculate the BIC criterion using the log likelihood and the # of
 **  clusters.
 **/  
double ClustEM::crisBIC( double logLike, int nClust ){
  // the # of free parameters to be estimated in this model is
  // 8 * nClust
  double bic = - 8*nClust;   
  bic = bic * log( nCris_ );
  bic = bic + 2*logLike;
  
  return( bic );

}

/**************************************************************************
 **  Apply the clustering approach outlined in Fraley and Raftery (2002).
 **/  
void ClustEM::fitClustModel(){

  // For the time being turn off the priors; these will be used only for
  // the final EM model-fitting.
  bool priorSpec = usingPrior;
  usingPrior = false;

  //   First, apply model-based hierarchical clustering,
  //   which yields a hierarchy of cluster assignments;
  // these are the rows of initClust.
  Array1D<double> *initClust = hierarClust();

  // For each choice of K, apply EM updating, not using the 
  // prior information, and evaluate the resulting BIC.  
  // The # of choices for K is the # of observations.
  int i, nClust, nChoices = nCris_;
  // Initialize the vector of the BIC values
  Array1D<double> crit (nChoices, 0.0);
  double logLike, maxBIC = -DBL_MAX;
  Array2D<double> clustAssign;

  for( i=0; i<nChoices; i++ ){  
    // For each value of K

    // apply EM updating of the cluster assigment,
    // where the # of clusters is fixed.
    clustAssign = emClust( initClust[i], logLike );

    // Calculate the BIC criterion for this cluster assignment
    nClust = clustAssign.dim2();
    crit[i] = crisBIC( logLike, nClust );

    cout << "BIC: " << crit[i] << endl;
    
    // If this gives a better value of BIC, update bestClust_
    if( crit[i] > maxBIC ){
      bestNclust_ = nClust;
      bestClust_ = clustAssign;
      maxBIC = crit[i];
    }

  }

  // Get
  // the probabilistic cluster assignment for the 
  // best choice of K by maximizing the posterior density
  if( priorSpec ){
    usingPrior = true;
    // apply EM updating of the cluster assigment,
    // where the # of clusters is fixed.
    clustAssign = emClust( initClust[nCris_-bestNclust_], logLike );
    bestClust_ = clustAssign;
  }
  
  modelFit = true;   // whether the model has been fit
  
}


/**************************************************************************
 **  Once the cluster model has been fit, get the best # of clusters
 **/  
int ClustEM::bestNclust(){	
  return (bestNclust_);
}

/**************************************************************************
 **  Once the cluster model has been fit, get the cluster assignment
 **  probabilities
 **/  
Array2D<double> ClustEM::bestClust(){
  return( bestClust_ );
}
