#ifndef CLUST_EM_H
#define CLUST_EM_H


#include "tnt_array1d.h"
#include "tnt_array2d.h"
#include "tnt_array1d_utils.h"
#include "tnt_array2d_utils.h"
#include "moreArray1dUtils.h"
#include "moreArray2dUtils.h"
#include "randGen.h"
#include "timeKeeper.h"
#include <cstdlib>
#include <assert.h>
#include <S.h>

using TNT::Array1D;
using TNT::Array2D;


class ClustEM{
 private:
  Array2D<double> *crisisDat_;   // The time series data for each crisis
  Array2D<double> **crisisCounts;    // The transition counts for each
				     // metric during each crisis.

  Array1D<double> firstPrior_;
  Array2D<double> transPrior_;
  int nCris_, nMetr_, bestNclust_;
  bool modelFit, usingPrior;
  Array2D<double> bestClust_;

  // These count objects are changed at each iteration of the MCMC
  Array1D<double> **firstCounts;
  Array2D<double> **transCounts;

  /**************************************************************************
   ** Order a vector of crisis type indicators (leave no gaps in indexing)
   **/  
  Array1D<double> orderTypeVec( const Array1D<double> &crisType );


  /**************************************************************************
   ** Calculate the counts for the crisis types and store in firstCounts and 
   ** transCounts
   **/  
  void calcCounts( const Array1D<double> &crisType ); 

  /**************************************************************************
   ** Calculate the counts for the crisis types, where the cluster
   ** assignments are probabilistic; store the resulting counts in
   ** firstCounts and transCounts.
   **/  
  void calcCounts( const Array2D<double> &clustAssign);

  /**************************************************************************
   **  given a cluster assignment, find the maximum classification
   **  log-likelihood.  clustAssign is the cluster assignment vector.
   **/
  double maxLogLike
  ( const Array1D<double> &clustAssign );

  /**************************************************************************
   ** calculate the log likelihood ratio for merging clusters clust1
   ** and clust2
   **/  
  double likeRatio
  ( const Array1D<double> &clustAssign, 
    int clust1, int clust2 );


  /**************************************************************************
   **  Apply model-based hierarchical clustering, which yields a
   **  hierarchy of cluster assignments; these are the rows of the
   **  returned matrix.
   **/
  Array1D<double> *hierarClust();


  /**************************************************************************
   **  Find cluster parameters that maximize the posterior density, for a
   **  particular cluster.  datClustPr is the current matrix of
   **  probabilistic assignments of observations to clusters, and
   **  clustInd is the index of the cluster of interest.  The MAPs are
   **  returned in mapGam and mapT.
   **/
  void mapParams
  ( int clustInd, Array1D<double> *mapGam, 
    Array2D<double> *mapT );

  /**************************************************************************
   **  Evaluate the likelihood for a single observation, given a set of
   **  parameter values.  obsInd is the index of the observation. 
   **/
  double evalLike( int obsInd,
      Array1D<double> *gam,
      Array2D<double> *T,
      bool logCalc );

  /**************************************************************************
   **  Calculate the BIC criterion using the log likelihood and the # of
   **  clusters.
   **/  
  double crisBIC( double logLike, int nClust );

  /**************************************************************************
   **    EM for clustering.  Requires an initial cluster assignment,
   **    given in the vector clusts.  Iterates until the improvement in
   **    log-likelihood is less than .01, i.e. until the likelihood
   **    changes by less than a factor of 1%.  Returns the
   **    log-likelihood in logLike.
   **/  
  Array2D<double> emClust
    ( const Array1D<double> &clusts, double &logLike );

 public:
  ClustEM(Array2D<double> *crisisList, int nCris,
	  const Array1D<double> &firstPrior,
	  const Array2D<double> &transPrior);

  ~ClustEM();

  /**************************************************************************
   **  Apply the clustering approach outlined in Fraley and Raftery (2002).
   **/  
  void fitClustModel();

  /**************************************************************************
   **  Once the cluster model has been fit, get the best # of clusters
   **/  
  int bestNclust();

  /**************************************************************************
   **  Once the cluster model has been fit, get the cluster assignment
   **  probabilities
   **/  
  Array2D<double> bestClust();

};

#endif
