
#include "CrisisMCMC.h"
//#include "randGen.h"

/* debugging info: 
   To print high-level debugging info at runtime, add "-DDEBUG1" to the first line of the Makevars
   file. 
*/



#define SQR(x) ((x)*(x))

using TNT::Array1D;
using TNT::Array2D;
using std::cout;
using std::endl;

extern "C"
{
  double S_STDCALL S_rpois( double lambda );
  double S_STDCALL S_rgamma( double s );
  double S_STDCALL S_runif( void );
  double S_STDCALL S_rexp( void );
}


void printAcceptProb( double logProb ){
  if (logProb > 1000)
    printf("log accept prob = infty \n" );
  else if (logProb < -1000)
    printf("log accept prob = - infty \n" );
  else 
    printf("log accept prob = %f\n", logProb );
}

CrisisMCMC::CrisisMCMC
(Array2D<double> *crisisList,
 int nCrises, double *typeInit,
 const Array1D<double> &firstPrior,
 const Array2D<double> &transPrior,
 double alpha,
 const Array1D<double> &temp,
 int nIter,
 int nBurnin, int nSave,
 char *logFile, char *auxParamFile,
 bool sampAux,
 bool printStats)
{
  crisisList_ = crisisList;
  nCrises_ = nCrises;
  nMetrics_ = crisisList[0].dim2();
  #ifdef DEBUG1
    cout << "Crisis list [0]" << crisisList[0] << endl;
    cout << "nMetrics: " << nMetrics_ << endl;
  #endif
  typeInit_ = typeInit;
  firstPrior_ = firstPrior.copy();
  transPrior_ = transPrior.copy();
  alpha_ = alpha;
  temp_ = temp;
  nIter_ = nIter;
  nBurnin_ = nBurnin;
  nSave_ = nSave;
  logFile_ = logFile;
  auxParamFile_ = auxParamFile;
  sampAux_ = sampAux;
  printStats_ = printStats;

  // Calculate a constant for use in likelihood calculations
  likeConst = nMetrics_ * lgamma( sum( firstPrior_ ) );  // lgamma is the log-gamma function
  int k1, k2;
  for( k1=0; k1<3; k1++ ){
    likeConst -= nMetrics_ * lgamma( firstPrior_[k1] );
    likeConst += nMetrics_ * lgamma( transPrior_[k1][0] + transPrior_[k1][1] + transPrior_[k1][2] );
    for( k2=0; k2<3; k2++ ){
      likeConst -= nMetrics_ * lgamma( transPrior_[k1][k2] );
    }
  }
  

  // Create a vector of equal probabilities for each crisis.
  // this is used in the split-merge move.
  crisProbs = Array1D<double>( nCrises_, 1/(double)nCrises_ );

  // Create the counts objects.  
  // Calculate the transition counts for each 
  // crisis and metric.
  firstCounts = new Array1D<double> * [ nCrises_ ];
  transCounts = new Array2D<double> * [ nCrises_ ];
  crisisCounts = new Array2D<double> * [ nCrises_ ];

  int i,j, crisLen, thisObs, nextObs;
  for( i=0; i<nCrises_; i++ ){
    // For each crisis
    // create the object vectors
    firstCounts[i] = new Array1D<double> [ nMetrics_ ];
    transCounts[i] = new Array2D<double> [ nMetrics_ ];
    crisisCounts[i] = new Array2D<double> [ nMetrics_ ];
    
    // get the # of time periods in the crisis
    crisLen = crisisList_[i].dim1();
    for( j=0; j<nMetrics_; j++ ){
      // For each metric
      // create the objects
      // Don't have to create firstCounts[i][j]; this is handled
      // before use of transCounts during likelihood calculation.
      transCounts[i][j] = Array2D<double>( 3, 3, 0.0 );
      crisisCounts[i][j] = Array2D<double>( 3, 3, 0.0 );

      // Calculate the transition counts for this crisis and metric
      for( k1=0; k1<(crisLen-1); k1++ ){
	// for each time period in the crisis except the last	
	// find the transition from time k1 to time k1+1
	thisObs = (int) crisisList_[i][k1][j];
	nextObs = (int) crisisList_[i][k1+1][j];
	// increment the appropriate count
        crisisCounts [i][j][thisObs][nextObs] = crisisCounts [i][j][thisObs][nextObs] + 1;
      }
      #ifdef DEBUG1
        cout << "Crisis counts [" << i << "][" << j << "]" << crisisCounts[i][j] << endl;
      #endif
    }
  }
}

CrisisMCMC::~CrisisMCMC(){}

/**************************************************************************
 ** Create a vector of probabilities from a vector of (unnormalized)
 ** log probabilities.
 **/ 
Array1D<double> probVec
( const Array1D<double> & logProb
){
  // subtract the max
  Array1D<double> pVec = logProb - TNT::max( logProb );
  // exponentiate
  pVec = exp( pVec );
  // normalize
  pVec = pVec / TNT::sum(pVec);

  return( pVec );
}

/**************************************************************************
 ** Given a matrix of values 0, 1, or 2, calculate the mode of each column.
 **/ 
Array1D<double> CrisisMCMC::colMode
( const Array2D<double> & dat
){
  int nCols = dat.dim2(), nRows= dat.dim1();
  Array1D<double> modeVec ( nCols, -1.0 ),
    countVec( 3, 0.0 );

  int i,j, mostCommon, thisVal;
  for(i=0; i<nCols; i++){
    // for each column
    // resent the count vector to 0s
    countVec = 0.0;
    for(j=0; j<nRows; j++){
      // for each row, increment the appropriate count
      thisVal = (int) dat[j][i];
      countVec[ thisVal ] = countVec[ thisVal ] + 1;
    }
    max( countVec, mostCommon );
    modeVec[i] = mostCommon;
  }
  return( modeVec );
}

/**************************************************************************
 ** Order a vector of crisis type indicators (leave no gaps in indexing)
 **/  
Array1D<double> CrisisMCMC::orderTypeVec( const Array1D<double> &crisType ){
  
  Array1D<double> crisLabels = unique( crisType );
  int i;
  Array1D<double> orderedTypes( nCrises_, 0.0 );
  for( i=0; i<nCrises_; i++ ){
    orderedTypes[i] = find( crisLabels, crisType[i] );
  }	
  return( orderedTypes );
}


/**************************************************************************
 ** Given a crisis type vector and an index i, find the number of
 ** possible types for crisis i.
 **/  
int CrisisMCMC::numPossibleTypes( const Array1D<double> &crisType, int i ){
  
  // find the # of possible types for crisis i, by setting crisType[i]
  // to a value that does not match any other types, and counting the 
  // resulting # of unique type values.
  Array1D<double> typeCopy = crisType.copy();
  typeCopy[i] = -1.0;
  int nPossTypes = unique( typeCopy ).dim();
      
  #ifdef DEBUG1
    cout << "Crisis " << i << ", num. possible types:" << nPossTypes << endl;
  #endif

  return( nPossTypes );
}

/**************************************************************************
 ** Calculate the counts for the crisis types and store in firstCounts and 
 ** transCounts
 **/  
void CrisisMCMC::calcCounts
( const Array1D<double> &crisType
){

  // Calculate the counts n_ijs where i is the crisis type, j is the metric,
  // and s is the value of the first observation in the crisis period.
  int i, j, numTypes = (int) max( crisType ) + 1;  // +1 due to 0 indexing

  for( i=0; i<numTypes; i++ ){
    // For each crisis type

    for( j=0; j<nMetrics_; j++ ){
      // For each metric:
      // start with the prior counts
      firstCounts[i][j] = firstPrior_.copy();
      transCounts[i][j] = transPrior_.copy();
    }
  }
  
  // Calculate the counts

  int thisCrisType, thisObs;
  for( i=0; i<nCrises_; i++ ){
    // for each crisis
    thisCrisType = (int) crisType[i];

    for( j=0; j<nMetrics_; j++ ){
      // for each metric
      // find the first value of this metric during this crisis
      thisObs = (int) crisisList_[i][0][j];
      // increment the appropriate first count
      firstCounts [thisCrisType][j][thisObs] = firstCounts [thisCrisType][j][thisObs] + 1;
      
      // add the transition counts for this crisis to the transition counts for this 
      // crisis type
      transCounts [thisCrisType][j] = transCounts [thisCrisType][j] + crisisCounts[i][j];
    }
  }
  #ifdef DEBUG_COUNTS
    cout << "crisis types: " << crisType << endl; 
    for( i=0; i<numTypes; i++ ){
      // For each crisis type

      for( j=0; j<nMetrics_; j++ ){
        // For each metric:
        cout << " firstCounts i=" << i << " j=" << j << endl; 
	cout << firstCounts[i][j] << endl;
        cout << " transCounts: " << endl; 
        cout << transCounts[i][j] << endl;
      }
    }
  #endif 

}

/**************************************************************************
 ** Calculate the marginal likelihood of a particular vector of crisis
 ** indicators.  Temper the likelihood by thisTemp.
 **/  
double CrisisMCMC::logLikelihood
( const Array1D<double> &crisType, 
  double thisTemp ){

  // time the likelihood calculation
  double startTime = tk.clockSecs();

  // Calculate the counts n_ijs where i is the crisis type, j is the metric,
  // and s is the value of the first observation in the crisis period.
  int i, j, k1, k2, numTypes = (int) max( crisType ) + 1;  // +1 due to 0 indexing

  calcCounts( crisType );

  // Calculate the marginal log-likelihood from the counts
  double *thisTrans;
  Array1D<double> transArr;
  // Calculate the part of the likelihood that involves the prior counts
  double logLike = numTypes * likeConst;

  for( i=0; i<numTypes; i++ ){
    // for each type
    for( j=0; j<nMetrics_; j++ ){
      // for each metric
      // The following line is faster than:  logLike -= lgamma( sum( firstCounts[i][j] ) );
      logLike -= lgamma( firstCounts[i][j][0] + firstCounts[i][j][1] + firstCounts[i][j][2] );
      for( k1=0; k1<3; k1++ ){
	// for each possible observed value
	logLike += lgamma( firstCounts[i][j][k1] );

	// Now handle the contribution of the transition counts to the 
	// marginal likelihood
	thisTrans = transCounts[i][j][k1];
        logLike -= lgamma( thisTrans[0] + thisTrans[1] + thisTrans[2] );
	for( k2=0; k2<3; k2++ ){
	  // for each possible observed transition
	  logLike += lgamma( thisTrans[k2] );
	}
      }
    }
  }
  #ifdef DEBUG_PROB
    cout << "log likelihood: " << logLike << endl;
  #endif 

  // temper by thisTemp
  logLike = logLike / thisTemp;

  // keep track of time spent on likelihood evals
  likeEvalTime += tk.clockSecs() - startTime;

  return(logLike);
}

/**************************************************************************
 ** Calculate the prior probability of a particular vector of crisis
 ** indicators
 **/  
double CrisisMCMC::logPriorProb( const Array1D<double> &crisType ){
  
  double logPrior = 0.0;

  // keep a tally of the number of crises so far seen
  // for each type
  int numTypes = (int) max( crisType ) + 1;  // +1 due to 0 indexing
  Array1D<int> crisCounts ( numTypes, 0 );  
  int i, thisType, numSame;

  // in the following computation, we have left out division
  // by (alpha_ + i), since it is the same for all crisType vectors.
  // (See Jain and Neal, 2004, p.164 for more details).
  // The following computation could be simplified by keeping track of 
  // the counts for each crisis type; the prior prob. is just the 
  // product of the gamma function evaluated at these counts, and could
  // be updated instead of recalculating each time.
  for( i=0; i<nCrises_; i++ ){
    // for each crisis

    thisType = (int) crisType[i];
    numSame = crisCounts[ thisType ];
    if( numSame <= 0 ){
      // if this crisis type has not yet been seen
      logPrior += log( alpha_ );
    } else {
      // if this crisis type has been seen
      logPrior += log( numSame );
    }
    // increment the counts vector
    crisCounts[ thisType ] = crisCounts[ thisType ] + 1;
  }

  #ifdef DEBUG_PROB
    cout << "log prior prob: " << logPrior << endl;
  #endif 
  return( logPrior );
}


/**************************************************************************
 ** Calculate the unnormalized posterior probability of a particular
 ** vector of crisis indicators.  Temper the likelihood by thisTemp.
 **/  
double CrisisMCMC::logPostProb
( const Array1D<double> &crisType,
  double thisTemp ){

  Array1D<double> orderedType = orderTypeVec( crisType );
  
  return( logLikelihood( orderedType, thisTemp ) + logPriorProb( orderedType ) );
}

/**************************************************************************
 **  Propose a split-merge move, and accept with the appropriate
 **  probability.  Update currTypes if appropriate.  Return 1 if the
 **  move is accepted, and 0 otherwise.
 **/  
int CrisisMCMC::splitMerge( Array1D<double> &currTypes, double thisTemp ){

  // sample the split / merge crises
  int i, ind1 = rdiscrete(crisProbs), ind2 = rdiscrete(crisProbs);
  if( ind1 == ind2 )
    return 0;

  double type1 = currTypes[ind1],
    type2 = currTypes[ind2];

  Array1D<double> newTypes = currTypes.copy();
  double accProb = 0;
  int nSplit = 0;
  if( type1 == type2 ){
    // split move
    double nextType = max(currTypes) + 1;
    newTypes[ind1] = nextType;
    for( i=0; i<nCrises_; i++ ){
      // for every crisis except ind2 & ind1
      if( (i != ind2) && (newTypes[i] == type1) ){
	// if it has type type1 then
	// with probability 0.5, change the type to nextType
	if( S_runif() < 0.5 )
	  newTypes[i] = nextType;
	// count the number of times that we "flip
	// this coin"
	nSplit++;
      }
    }
    // calculate the merge proposal prob. / split proposal prob.
    accProb = nSplit * log(2.0);
    #ifdef DEBUG_SM
      cout << "Split proposed.  ind1: " << ind1 << "  ind2: " << ind2 <<
	"  type1:" << type1 << "  type2:" << type2 << "  nSplit: " << 
	nSplit << "  New types: " << newTypes << endl;
    #endif
    nTrySplit++;

  } else {
    // merge move
    for( i=0; i<nCrises_; i++ ){
      // for every crisis
      if( currTypes[i] == type1 ){
	// if it has the type type1 then
	// change the type to type2
	newTypes[i] = type2;
      }
      if( ( currTypes[i] == type1 ) || ( currTypes[i] == type2 ) ){
	// count the number of crises of either type
	nSplit++;
      }
      accProb = (nSplit - 2) * log(0.5);
    }
    #ifdef DEBUG_SM
      cout << "Merge proposed.  ind1: " << ind1 << "  ind2: " << ind2 <<
	"  type1:" << type1 << "  type2:" << type2 << "  nSplit: " << 
	nSplit << "  New types: " << newTypes << endl;
    #endif

    nTryMerge++;
  }
  // multiply the acceptance probability by the tempered posterior ratio
  accProb += logPostProb( newTypes, thisTemp ) - logPostProb( currTypes, thisTemp );
  #ifdef DEBUG_SM
    cout << "S/M log acceptance prob: " << accProb << endl;
  #endif


  double samp = S_rexp();
  if( accProb + samp > 0 ) {   // accept the proposed split / merge
    currTypes = orderTypeVec( newTypes );
    #ifdef DEBUG_SM
      cout << "Move accepted " << endl;
    #endif
    if( type1 == type2 )
      nAccSplit++;
    else 
      nAccMerge++;
    return(1);
  } else {
    #ifdef DEBUG_SM
      cout << "Move rejected " << endl;
    #endif
    return(0);
  }
}

/**************************************************************************
 **  Sample the auxiliary parameters, and save to file.
 **/  

void CrisisMCMC::sampSaveAux
( Array1D<double> *crisType,
  FILE *auxParamFile
){

  int i, j, k1, k2;
  double nTypes;

  Array1D<double> beta, Trow, rowCounts;
  // find the number of types for the first temperature
  nTypes = max( crisType[0] ) + 1.0;
  // calculate the counts for all types
  calcCounts ( crisType[0] );

  #ifdef DEBUG_PRED
    cout << "crisType: " << crisType[0] << endl;
  #endif

  // save the number of crisis types
  // Write in binary format to save on write and read time & space
  fwrite( &nTypes, sizeof(double), 1, auxParamFile );
  
  for( i=0; i<(int)nTypes; i++ ){
    // for each crisis type
    for( j=0; j<nMetrics_; j++ ){
      // for each metric
      // sample beta
      beta = rdirichlet( firstCounts[i][j] );
      for( k1=0; k1< 3; k1++ ){
	// store beta
	fwrite( &(beta[k1]), sizeof(double), 1, auxParamFile );  
      }
      #ifdef DEBUG_PRED
      cout << "type: " << i << "  metric: " << j << "  firstCounts: " 
	   << firstCounts[i][j] << endl;
      cout << "beta: " << beta << endl;
      #endif
      for( k1=0; k1< 3; k1++ ){
	// for each row of the transition matrix
	rowCounts = Array1D<double> ( 3, transCounts[i][j][k1] );
	// sample this row
	Trow = rdirichlet( rowCounts );
	for( k2=0; k2< 3; k2++ ){
	  // store row of trans. matrix
	  fwrite( &(Trow[k2]), sizeof(double), 1, auxParamFile );  
	}
        #ifdef DEBUG_PRED
        cout << "row: " << k1 << "  rowCounts: " 
	     << rowCounts << endl;
        cout << "Trow: " << Trow << endl;
        #endif

      }
    }
  }
  
}

/**************************************************************************
 **  Run the MCMC
 **/  
void CrisisMCMC::simulate(){

  if (nSave_ == 0) {
    printf("nSave specified as zero, which is invalid.\n");
    return;
  }  

  //  Calculate the interval between state saves 
  int dSav = (int) floor( nIter_ / (double) nSave_ );  
  if (dSav == 0)   dSav = 1;

  // save the start time & initialize the likelihood evaluation time.
  double startTime = tk.clockSecs();
  likeEvalTime = 0.0;

  //  Initialize the crisis type indicators.  There is one
  // vector of crisis types for each temperature, giving 
  // an nTemps x nCrisis matrix.  Initialize the crisis
  // types for every temperature to typeInit_.
  int i,j, t, nTemps = temp_.dim();
  Array1D<double> *crisType = new Array1D<double> [nTemps];
  for( i=0; i<nTemps; i++ ){
    crisType[i] = Array1D<double>(nCrises_, 0.0);
    for( j=0; j<nCrises_; j++ )
      crisType[i][j] = typeInit_[j];
    // Correct the ordering of the initial types, if necessary.
    crisType[i] = orderTypeVec( crisType[i] );
    #ifdef DEBUG1
      cout << "Initial crisis type vector for temperature index " << i << ": " << crisType[i] << endl;
    #endif
  }


  //  Open the output file
  FILE *logFile = fopen(logFile_, "w" );
  FILE *auxParamFile = NULL;
  if( sampAux_ ){
    auxParamFile = fopen(auxParamFile_, "w" );
  }

  // Create a vector to hold the log probabilities
  // of each type of crisis.  
  Array1D<double> typeProbs;

  // create a vector of the probability of updating
  // each temperature, and one for the probability
  // of swapping each temperature with the next.
  Array1D<double> tempProbs (nTemps, 1/(double)nTemps),
    swapProbs;
  if( nTemps > 1 ){
    swapProbs = Array1D<double>( nTemps-1, 1/(double)(nTemps-1) );
  }
  #ifdef DEBUG1
    cout << "tempProbs:" << tempProbs << endl;
    cout << "swapProbs:" << swapProbs << endl;
  #endif

  // create a vector to hold the number of tries and acceptances for swaps
  // between each pair of adjacent temperatures.
  Array1D<double> nTrySwap( nTemps - 1, 0.0 ), nAccSwap( nTemps - 1, 0.0 );
  // initialize the counts for the number of tries and acceptances for
  // split-merge moves.
  nTrySplit=0;  nAccSplit = 0;  nTryMerge=0;  nAccMerge=0;

  // Create other objects needed during the MC updating
  int nPossTypes;
  double swapAccProb, samp;
  Array1D<double> swapVec;

  int itn = (-nBurnin_-1);
  while( itn < nIter_ ){
    // for each iteration 

    // choose a temperature to update
    t = rdiscrete(tempProbs);

    // update the crisis type vector for this temperature
    for( i=0; i<nCrises_; i++ ){
      // for each crisis, update the crisis type indicator for temp t

      #ifdef DEBUG1
        cout << "Current crisis type vector for temp " << temp_[t] << ": " << crisType[t] << endl;
      #endif

      // Calculate the number of possible types for crisis i.
      // Then the possible values for crisis i are 0, 1, ..., nPossTypes.
      // This can be seen by examining the cases where crisis i 
      // currently has or does not have unique type.
      nPossTypes = numPossibleTypes( crisType[t], i );

      // Calculate the log probability of each crisis type.
      // First set the type prob. vector to zeros.
      typeProbs = Array1D<double> ( nPossTypes, 0.0 );

      for( j=0; j<nPossTypes; j++ ){
	// for each possible type
	// calculate the unnormalized log post. prob. for
	// the crisis type vector with this type assignment.
	crisType[t][i] = j;
	typeProbs[j] = logPostProb( crisType[t], temp_[t] );
      }
      #ifdef DEBUG1
        cout << "type log probs:" << typeProbs << endl;
      #endif
      // Transform to find probabilities.  It should 
	// be possible to avoid this 
      typeProbs = probVec( typeProbs );
      #ifdef DEBUG1
        cout << "type probs:" << typeProbs << endl;
      #endif
      // Sample the type indicator for this crisis
      crisType[t][i] = (double) rdiscrete( typeProbs );
      // Relabel the crises to obtain the correct ordering.
      crisType[t] = orderTypeVec( crisType[t] );

      #ifdef DEBUG1
        cout << "Updated crisis type vector for temp " << temp_[t] << ": " << crisType[t] << endl;
      #endif
    }

    
    // Attempt split-merge moves.
    for( i=0; i<nCrises_; i++){
      // choose a temperature to update
      t = rdiscrete(tempProbs);
      #ifdef DEBUG_SM
        cout << "Vector for temp " << temp_[t] << " before split-merge: " << crisType[t] << endl;
      #endif
  
      splitMerge( crisType[t], temp_[t] );
  
      #ifdef DEBUG_SM
        cout << "Vector for temp " << temp_[t] << " after split-merge: " << crisType[t] << endl;
      #endif
    }

    // Attempt a swap move
    if( nTemps > 1 ){
      // if there is more than one temperature, choose a temp. to swap
      t = rdiscrete( swapProbs );
      // calculate the log acceptance probability
      swapAccProb = logLikelihood( crisType[t], temp_[t+1] );
      swapAccProb = swapAccProb + logLikelihood( crisType[t+1], temp_[t] );
      swapAccProb = swapAccProb - logLikelihood( crisType[t], temp_[t] );
      swapAccProb = swapAccProb - logLikelihood( crisType[t+1], temp_[t+1] );
      // swap with this (log) probability	
      samp = S_rexp();
      if( swapAccProb + samp > 0 ) {   // accept the proposed swap
	swapVec = crisType[t].copy();
	crisType[t] = crisType[t+1].copy();
	crisType[t+1] = swapVec.copy();
	nAccSwap[t]++;
      }
      nTrySwap[t]++;
    }

    if( ((itn % dSav) == 0) && (itn >= 0) ){
      // ---------------------   Write state to files  ------------------------
      // Save the crisis vectors corresponding to each of the temperatures.
      for( t=0; t<nTemps; t++ ){
	for( i=0; i<nCrises_; i++ ){
  	  // Write in binary format to save on write and read time & space
	  fwrite( &crisType[t][i], sizeof(double), 1, logFile );  
	}
      }
      if( sampAux_ ){
        sampSaveAux( crisType, auxParamFile );
      }
    }

    itn++;
  }



  // --------------------------   Wrap Up  -----------------------------
  //  Close the output files
  fclose( logFile );
  if( sampAux_ ){
    fclose( auxParamFile );
  }
  //  Print the MCMC statistics if requested.
  if((itn>0) && (printStats_)){
    double compTime = tk.clockSecs() - startTime;
    printf( "Total computation time: %f\n", compTime );
    printf( "Likelihood evaluation time: %f\n", likeEvalTime );
    std::cout << "Swaps attempted: " << nTrySwap << "and accepted: " << nAccSwap << std::endl;
    std::cout << "Splits attempted: " << nTrySplit << "  and accepted: " << nAccSplit << std::endl;
    std::cout << "Merges attempted: " << nTryMerge << "  and accepted: " << nAccMerge << std::endl;
  }
  
}


