#ifndef CRISIS_MCMC_H
#define CRISIS_MCMC_H


#include "tnt_array1d.h"
#include "tnt_array2d.h"
#include "tnt_array1d_utils.h"
#include "tnt_array2d_utils.h"
#include "moreArray1dUtils.h"
#include "moreArray2dUtils.h"
#include "randGen.h"
#include "timeKeeper.h"
#include <cstdlib>
#include <assert.h>
#include <S.h>

using TNT::Array1D;
using TNT::Array2D;


class CrisisMCMC{
 private:
  Array2D<double> *crisisList_;   // The time series data for each crisis
  Array2D<double> **crisisCounts;    // The transition counts for each
				     // metric during each crisis.

  Array1D<double> firstPrior_, temp_, crisProbs;
  Array2D<double> transPrior_;
  double *typeInit_;
  double alpha_, likeConst, likeEvalTime;
  int nCrises_, nMetrics_, nIter_, nBurnin_, nSave_,
    nTrySplit, nAccSplit, nTryMerge, nAccMerge;
  char *logFile_, *auxParamFile_;
  bool sampAux_, printStats_;
  timeKeeper tk;

  // These count objects are changed at each iteration of the MCMC
  Array1D<double> **firstCounts;
  Array2D<double> **transCounts;

  /**************************************************************************
   ** Given a matrix of values 0, 1, or 2, calculate the mode of each column.
   **/  
  Array1D<double> colMode ( const Array2D<double> & dat );

  /**************************************************************************
   ** Order a vector of crisis type indicators (leave no gaps in indexing)
   **/  
  Array1D<double> orderTypeVec( const Array1D<double> &crisType );


  /**************************************************************************
   ** Given a crisis type vector and an index i, find the number of
   ** possible types for crisis i.
   **/  
  int numPossibleTypes( const Array1D<double> &crisType, int i );
   
  /**************************************************************************
   ** Calculate the counts for the crisis types and store in firstCounts and 
   ** transCounts
   **/  
  void calcCounts( const Array1D<double> &crisType ); 

  /**************************************************************************
   ** Calculate the marginal likelihood of a particular vector of
   ** crisis indicators.  crisType should be correctly ordered, such
   ** as the result of a call to orderTypeVec.  Temper the likelihood
   ** by thisTemp.
   **/  
  double logLikelihood( const Array1D<double> &crisType, double thisTemp );

  /**************************************************************************
   ** Calculate the prior probability of a particular vector of crisis
   ** indicators.  crisType should be correctly ordered, such as the
   ** result of a call to orderTypeVec.
   **/  
  double logPriorProb( const Array1D<double> &crisType );

  /**************************************************************************
   ** Calculate the log of the unnormalized posterior probability of a
   ** particular vector of crisis indicators.  crisType should be
   ** correctly ordered, such as the result of a call to
   ** orderTypeVec. Temper the likelihood by thisTemp.
   **/  
  double logPostProb( const Array1D<double> &crisType, double thisTemp );

  /**************************************************************************
   **  Propose a split-merge move, and accept with the appropriate
   **  probability.  Update currTypes if appropriate.  Return 1 if the
   **  move is accepted, and 0 otherwise.
   **/  
  int splitMerge( Array1D<double> &currTypes, double thisTemp );

  /**************************************************************************
   **  Sample the auxiliary parameters, and save to file.
   **/  
  void sampSaveAux( Array1D<double> *crisType,
		    FILE *auxParamFile );  

 public:
  CrisisMCMC(Array2D<double> *crisisList,
	     int nCrises, double *typeInit,
	     const Array1D<double> &firstPrior,
	     const Array2D<double> &transPrior,
	     double alpha, 
	     const Array1D<double> &temp,
	     int nIter, int nBurnin, 
	     int nSave,
	     char *logFile, char *auxParamFile,
	     bool sampAux,
	     bool printStats = true);         // Whether to print summary
				      // statistics about the MCMC
				      // (e.g. acceptance rates)

  ~CrisisMCMC();

  /**************************************************************************
   **  Run the MCMC
   **/  
  void simulate();

};

#endif
