
#include "PredCrisis.h"
//#include "randGen.h"

/* debugging info: 
   To print debugging info at runtime, add "-DDEBUG_PRED" to the first line of the Makevars
   file. 
*/



#define SQR(x) ((x)*(x))

using TNT::Array1D;
using TNT::Array2D;
using std::cout;
using std::endl;

extern "C"
{
  double S_STDCALL S_rpois( double lambda );
  double S_STDCALL S_rgamma( double s );
  double S_STDCALL S_runif( void );
  double S_STDCALL S_rexp( void );
}


PredCrisis::PredCrisis
(const Array2D<double> &newDat,
 Array2D<double> *crisisList,
 char *mcFile,
 int nCrises, int nMetr,
 int nTemps, double alpha,
 const Array1D<double> &firstPrior,
 const Array2D<double> &transPrior,
 bool printInfo)
{
  newDat_ = newDat.copy();
  crisisList_ = crisisList;
  mcFile_ = mcFile;
  nCris_ = nCrises; 
  nMetr_ = nMetr;
  nTemps_ = nTemps;
  alpha_ = alpha;
  firstPrior_ = firstPrior.copy();
  transPrior_ = transPrior.copy();
  printInfo_ = printInfo;

  // calculate a useful constant for marginal likelihood calculations
  int k1,k2;
  likeConst = nMetr_ * lgamma( sum( firstPrior_ ) );  // lgamma is the log-gamma function
  for( k1=0; k1<3; k1++ ){
    likeConst -= nMetr_ * lgamma( firstPrior_[k1] );
    likeConst += nMetr_ * lgamma( transPrior_[k1][0] + transPrior_[k1][1] + transPrior_[k1][2] );
    for( k2=0; k2<3; k2++ ){
      likeConst -= nMetr_ * lgamma( transPrior_[k1][k2] );
    }
  }
  #ifdef DEBUG_PRED
    cout << "Likelihood constant: " << likeConst << endl;
  #endif

  // Create the counts objects.
  // Calculate the transition counts for each 
  // previous crisis and each metric.
  firstCounts = new Array1D<double> * [ nCris_ + 1 ];
  transCounts = new Array2D<double> * [ nCris_ + 1 ];
  crisisCounts = new Array2D<double> * [ nCris_ + 1 ];

  int i,j, crisLen, thisObs, nextObs;
  for( i=0; i<nCris_; i++ ){
    // For each crisis
    // create the object vectors
    firstCounts[i] = new Array1D<double> [ nMetr_ ];
    transCounts[i] = new Array2D<double> [ nMetr_ ];
    crisisCounts[i] = new Array2D<double> [ nMetr_ ];
    
    // get the # of time periods in the crisis
    crisLen = crisisList_[i].dim1();
    for( j=0; j<nMetr_; j++ ){
      // For each metric
      // create the objects
      // Don't have to create firstCounts[i][j]; this is handled
      // before use of transCounts during likelihood calculation.
      transCounts[i][j] = Array2D<double>( 3, 3, 0.0 );
      crisisCounts[i][j] = Array2D<double>( 3, 3, 0.0 );

      // Calculate the transition counts for this crisis and metric
      for( k1=0; k1<(crisLen-1); k1++ ){
	// for each time period in the crisis except the last	
	// find the transition from time k1 to time k1+1
	thisObs = (int) crisisList_[i][k1][j];
	nextObs = (int) crisisList_[i][k1+1][j];
	// increment the appropriate count
        crisisCounts [i][j][thisObs][nextObs] = crisisCounts [i][j][thisObs][nextObs] + 1;
      }
      #ifdef DEBUG1
        cout << "Crisis counts [" << i << "][" << j << "]" << crisisCounts[i][j] << endl;
      #endif
    }
  }

  // calculate the transition counts for the new data
  // create the object vectors
  firstCounts[nCris_] = new Array1D<double> [ nMetr_ ];
  transCounts[nCris_] = new Array2D<double> [ nMetr_ ];
  crisisCounts[nCris_] = new Array2D<double> [ nMetr_ ];

  crisLen = newDat_.dim1();
  for( i=0; i<nMetr_; i++ ){
    // for each metric
    //create the counts objects
    transCounts[nCris_][i] = Array2D<double>( 3, 3, 0.0 );
    crisisCounts[nCris_][i] = Array2D<double> (3, 3, 0.0);
    for( k1=0; k1<(crisLen-1); k1++ ){
      // for each time period in the crisis except the last	
      // find the transition from time k1 to time k1+1
      thisObs = (int) newDat_[k1][i];
      nextObs = (int) newDat_[k1+1][i];
      // increment the appropriate count
      crisisCounts [nCris_][i][thisObs][nextObs] = crisisCounts [nCris_][i][thisObs][nextObs] + 1;
    }
    #ifdef DEBUG_PRED
      cout << "new data transition counts for metric " << i << ": " << crisisCounts[nCris_][i] << endl;
    #endif
  }

}

PredCrisis::~PredCrisis(){}

/**************************************************************************
 ** Create a vector of probabilities from a vector of (unnormalized)
 ** log probabilities.
 **/ 
Array1D<double> PredCrisis::toProbs
( const Array1D<double> & logProb
){
  // subtract the max
  Array1D<double> pVec = logProb - TNT::max( logProb );
  // exponentiate
  pVec = exp( pVec );
  // normalize
  pVec = pVec / TNT::sum(pVec);

  return( pVec );
}


/**************************************************************************
 ** Calculate the counts for the crisis types and store in firstCounts
 ** and transCounts.  firstCounts consists of the counts n_ijs where i
 ** is the crisis type, j is the metric, and s is the value of the
 ** first observation in the crisis period.  transCounts is similarly
 ** the count of transitions from s to t for each type and metric.
 **/  
void PredCrisis::calcCounts
( const Array1D<double> &crisType
){

  int i, j, numTypes = (int) max( crisType ) + 1;  // +1 due to 0 indexing

  for( i=0; i<numTypes; i++ ){
    // For each crisis type

    for( j=0; j<nMetr_; j++ ){
      // For each metric:
      // start with the prior counts
      firstCounts[i][j] = firstPrior_.copy();
      transCounts[i][j] = transPrior_.copy();
    }
  }
  
  // Calculate the counts

  int thisCrisType, thisObs;
  for( i=0; i<(nCris_+1); i++ ){
    // for each crisis, including the new one
    thisCrisType = (int) crisType[i];

    for( j=0; j<nMetr_; j++ ){
      // for each metric
      // find the first value of this metric during this crisis
      if( i < nCris_ )
        thisObs = (int) crisisList_[i][0][j];
      else
	thisObs = (int) newDat_[0][j];
      // increment the appropriate first count
      firstCounts [thisCrisType][j][thisObs] = firstCounts [thisCrisType][j][thisObs] + 1;
      
      // add the transition counts for this crisis to the transition counts for this 
      // crisis type
      transCounts [thisCrisType][j] = transCounts [thisCrisType][j] + crisisCounts[i][j];
    }
  }
  #ifdef DEBUG_COUNTS
    cout << "crisis types: " << crisType << endl; 
    for( i=0; i<numTypes; i++ ){
      // For each crisis type

      for( j=0; j<nMetr_; j++ ){
        // For each metric:
        cout << " firstCounts i=" << i << " j=" << j << endl; 
	cout << firstCounts[i][j] << endl;
        cout << " transCounts: " << endl; 
        cout << transCounts[i][j] << endl;
      }
    }
  #endif 

}

/**************************************************************************
 ** Calculate the marginal likelihood of a particular vector of crisis
 ** indicators.  Temper the likelihood by thisTemp.
 **/  
double PredCrisis::logLikelihood
( const Array1D<double> &crisType, 
  double thisTemp ){

  int i, j, k1, k2, numTypes = (int) max( crisType ) + 1;  // +1 due to 0 indexing

  // Calculate the counts firstCounts and transCounts
  calcCounts( crisType );

  // Calculate the marginal log-likelihood from the counts
  double *thisTrans;
  Array1D<double> transArr;
  // Calculate the part of the likelihood that involves the prior counts
  double logLike = numTypes * likeConst;

  for( i=0; i<numTypes; i++ ){
    // for each type
    for( j=0; j<nMetr_; j++ ){
      // for each metric
      logLike -= lgamma( firstCounts[i][j][0] + firstCounts[i][j][1] + firstCounts[i][j][2] );
      for( k1=0; k1<3; k1++ ){
	// for each possible observed value
	logLike += lgamma( firstCounts[i][j][k1] );

	// Now handle the contribution of the transition counts to the 
	// marginal likelihood
	thisTrans = transCounts[i][j][k1];
        logLike -= lgamma( thisTrans[0] + thisTrans[1] + thisTrans[2] );
	for( k2=0; k2<3; k2++ ){
	  // for each possible observed transition
	  logLike += lgamma( thisTrans[k2] );
	}
      }
    }
  }
  #ifdef DEBUG_PROB
    cout << "log likelihood: " << logLike << endl;
  #endif 

  // temper by thisTemp
  logLike = logLike / thisTemp;

  return(logLike);
}

/**************************************************************************
 ** Perform prediction.  Return a vector of length # previous crises +
 ** 1, that contains the probability that the new crisis is of the
 ** same type as the previous crises (the last entry is the
 ** probability that the new crisis is of a new type)
 **/  
Array1D<double> PredCrisis::predict(){
  

  // open file containing MC samples
  FILE *mcFile = fopen(mcFile_, "r" );

  int i, j, res;
  Array1D<double> probVec (nCris_+1, 0.0),
    typeProbs, crisCounts, crisTypes(nCris_+1, 0.0),
    probSum( nCris_+1, 0.0 );

  int nTypes, nSamps = 0;
  double fromFile;
  while( fread( &fromFile, sizeof(double), 1, mcFile ) == 1 ){
    // for each posterior sample

    // read in crisis types 
    // The first has already been read in
    crisTypes[0] = fromFile;
    for( i=1; i<nCris_; i++ ){
      // for the rest of the crises
      // read its type from the file (for the first temperature)
      res = fread( &fromFile, sizeof(double), 1, mcFile );
      if( res != 1 )
	MESSAGE "File ended before content read\n" ERROR;

      crisTypes[i] = fromFile;
    }
    // calculate the number of types
    crisTypes[nCris_] = -1.0;
    nTypes = (int) TNT::max( crisTypes ) + 1;
    #ifdef DEBUG_PRED
      cout << "nTypes: " << nTypes << endl;
    #endif

    // Calculate the counts for each type of crisis
    // Create the counts object.
    crisCounts = Array1D<double> (nTypes, 0.0);
    // calculate the counts 
    for( i=0; i<nCris_; i++ ){
      crisCounts[ (int)crisTypes[i] ]++;      
    }
    #ifdef DEBUG_PRED
      cout << "crisis types: " << crisTypes << endl;
      cout << "crisis type counts: " << crisCounts << endl;
    #endif

    // Calculate the log probability that the new crisis is 
    // of each crisis type.
    // First set the type prob. vector to zeros.  The
    // last entry of this vector will be the prob. of
    // a new type.
    typeProbs = Array1D<double> ( nTypes+1, 0.0 );
    // Calculate the log prior probability of each type,
    // and add to the log likelihood
    for( i=0; i<nTypes; i++ ){
      // first the prior
      typeProbs[i] = log( crisCounts[i] );
      // now the likelihood
      crisTypes[nCris_] = i;  
      // The crisis type vector still has the correct ordering,
      // which is needed for the following call to logLikelihood
      typeProbs[i] += logLikelihood( crisTypes );
    }
    // handle the case of a new type
    // first the prior
    typeProbs[nTypes] = log( alpha_ );
    // now the likelihood
    crisTypes[nCris_] = nTypes;
    typeProbs[nTypes] += logLikelihood( crisTypes );

    // normalize typeProbs
    typeProbs = toProbs( typeProbs );

    #ifdef DEBUG_PRED
      cout << "final typeProbs vector: " << typeProbs << endl;
    #endif

    // for each past crisis, calculate pr(znew = zi)
    for( i=0; i<nCris_; i++ ){
      probVec[i] = typeProbs[ (int)crisTypes[i] ];
    }
    probVec[nCris_] = typeProbs[nTypes];

    #ifdef DEBUG_PRED
      cout << "probVec: " << probVec << endl;
    #endif

    // read through the end of this sample in the MC file
    // (all the temperatures are saved in this file, while
    // we are only predicting using the first temperature)
    for( j=1; j<nTemps_; j++ ){
      // for each temperature except the first
      for( i=0; i<nCris_; i++ ){
	// for each crisis
	res = fread( &fromFile, sizeof(double), 1, mcFile );
	if( res != 1 )
	  MESSAGE "File ended before content read\n" ERROR;

      }
    }
    probSum = probSum + probVec;
    nSamps++;
  }
  probVec = probSum / (double)nSamps;
  if( printInfo_ ){
    cout << "num. posterior samples: " << nSamps << endl;
  }

  fclose( mcFile ); 
    
  return( probVec );
}
