/*
*
* Template Numerical Toolkit (TNT)
*
* Mathematical and Computational Sciences Division
* National Institute of Technology,
* Gaithersburg, MD USA
*
*
* This software was developed at the National Institute of Standards and
* Technology (NIST) by employees of the Federal Government in the course
* of their official duties. Pursuant to title 17 Section 105 of the
* United States Code, this software is not subject to copyright protection
* and is in the public domain. NIST assumes no responsibility whatsoever for
* its use by other parties, and makes no guarantees, expressed or implied,
* about its quality, reliability, or any other characteristic.
*
*/

#include <cassert>
#include <fstream>
#include <cmath>
#include <vector>
#include "moreArray2dUtils.h"
#include "moreArray1dUtils.h"
#include "tnt_array2d_utils.h"
//#include "jama_qr.h"

using TNT::Array1D;
using TNT::Array2D;
using std::vector;
using std::cerr;
using std::endl;

/* Multiply an array by a scalar */
Array2D<double> TNT::operator*(const Array2D<double> &A, double b)
{
	int m = A.dim1();
	int n = A.dim2();

	Array2D<double> C(m,n);

	for (int i=0; i<m; i++)
	  {
	    for (int j=0; j<n; j++)
	      C[i][j] = A[i][j] * b;
	  }
	return C;
 }	

Array2D<double> TNT::operator*(double a, const Array2D<double> &B)
{
  return(B*a);
 }	

/* Divide an array by a scalar */
Array2D<double> TNT::operator/(const Array2D<double> &A, double b){
  return(A * pow(b, -1));
}


/* Compare each element of the array A to the constant b.  Equality is
  indicated by 1.0, while inequality is indicated by 0.0 */
Array2D<double> TNT::operator== (const Array2D<double> &A, double b){
  int i,j, 
    m = A.dim1(), n = A.dim2();

  Array2D<double> C(n,m);

  for (i=0; i<m; i++){
    for (j=0; j<n; j++)
      if( A[i][j] == b )
	C[i][j] = 1.0;
      else	
	C[i][j] = 0.0;
  }
  
  return C;
}


/* Matrix transpose */
Array2D<double> TNT::arrayTranspose(const Array2D<double> &A)
{
	int m = A.dim1();
	int n = A.dim2();

	Array2D<double> C(n,m);

	for (int i=0; i<m; i++)
	  {
	    for (int j=0; j<n; j++)
	      C[i][j] = A[j][i];
	  }
	return C;
 }	

/* column concatenation of matrices.  If the arrays do not have the
   same number of rows then returns a null array */
Array2D<double> TNT::cbind (const Array2D<double> &A, const Array2D<double> &B)
{
  int row,col, nRowsA = A.dim1(), nRowsB = B.dim1();
  int nColsA = A.dim2(), nColsB = B.dim2();

  if (nRowsA != nRowsB)
    return Array2D<double>();

  Array2D<double> boundMat (nRowsA, nColsA + nColsB );
  for (row=0; row<nRowsA; row++){
    for (col=0; col<nColsA; col++){
      boundMat[row][col] = A[row][col];
    }
    for (col=0; col<nColsB; col++){
      boundMat[row][col+nColsA] = B[row][col];
    }
  }
  return boundMat;

}


/* row concatenation of matrices.  If the arrays do not have the
   same number of columns then returns a null array */
Array2D<double> TNT::rbind (const Array2D<double> &A, const Array2D<double> &B)
{
  int row,col, nRowsA = A.dim1(), nRowsB = B.dim1();
  int nColsA = A.dim2(), nColsB = B.dim2();

  if (nColsA != nColsB)   //if the matrices do not have the same number of columns
    return Array2D<double>();

  Array2D<double> boundMat (nRowsA + nRowsB, nColsA );
  for (col=0; col<nColsA; col++){
    for (row=0; row<nRowsA; row++){
      boundMat[row][col] = A[row][col];
    }
    for (row=0; row<nRowsB; row++){
      boundMat[row+nRowsA][col] = B[row][col];
    }
  }		
  return boundMat;


}

/* the nrows x nrows identity matrix */
Array2D<double> TNT::identity (int nrows){
  int i;
  Array2D<double> id (nrows, nrows, 0.0);
  
  for (i=0; i<nrows; i++)
    id[i][i] =1;

  return id;
}

/* the nrows x ncols matrix where every i,i element is 1 and the other
   elements are 0. */
Array2D<double> TNT::identity (int nrows, int ncols){
  Array2D<double> id (nrows, ncols, 0.0);
  
  int n1elements = std::min (nrows, ncols);
  for (int i=0; i<n1elements; i++)
    id[i][i] =1;

  return id;
}

/* Obtain the column sums of the matrix A. */
Array1D<double> TNT::colSums (const Array2D<double> &A){

  int i,j, nrow = A.dim1(), ncol = A.dim2();
  Array1D<double> res( ncol, 0.0 ); 
  for( i=0; i<nrow; i++ ){
    for( j=0; j<ncol; j++ ){
      res[j] += A[i][j];
    }
  }
  return(res);
}


/* the nrows x nrows matrix where the i,i element is elems[i] and the other
   elements are 0. */
Array2D<double> TNT::diag (const Array1D<double> &elems){
  int nrows = elems.dim();
  Array2D<double> diagonal (nrows, nrows, 0.0);

  for (int i=0; i<nrows; i++)
    diagonal[i][i] =elems[i];

  return diagonal;			    
}


/* the subarray defined by the specified rows and cols of A */
Array2D<double> TNT::subarray (const Array2D<double> &A, int minRow, int maxRow,
			   int minCol, int maxCol){

   int row, col, nRows = maxRow - minRow + 1,
     nCols = maxCol - minCol + 1;

  if (nRows <= 0)
    return Array2D<double> (0, nCols);

  if (nCols <= 0)
    return Array2D<double> (nRows, 0);

  assert ((minRow >= 0) && (maxRow < nRows) &&
	  (minCol >= 0) && (maxCol < nCols));

   Array2D<double> B (nRows, nCols);

   for (row = minRow; row <= maxRow; row++)
     for (col = minCol; col <= maxCol; col++)
       B [row - minRow][col-minCol] = A [row] [col];

   return B;
 }

/* the subarray defined by removing the row corresponding to rowToRemove */
Array2D<double> TNT::removeRow (const Array2D<double> &A, int rowToRemove){
  
  int ncols = A.dim2(), nrows = A.dim1();
  assert ((rowToRemove < nrows) && (rowToRemove >= 0));

  Array2D<double> topSubarray = TNT::subarray (A, 0, rowToRemove-1, 0, ncols),
    bottomSubarray = TNT::subarray (A, rowToRemove+1, nrows, 0, ncols),
    rowRemovedArray = TNT::rbind(topSubarray, bottomSubarray);

  return rowRemovedArray;
}

/* the subarray defined by removing the row corresponding to rowToRemove */
Array2D<double> TNT::removeCol (const Array2D<double> &A, int colToRemove){

  int ncols = A.dim2(), nrows = A.dim1();

  assert ((colToRemove < ncols) && (colToRemove >= 0));
  Array2D<double> leftSubarray = TNT::subarray (A, 0, nrows, 0, colToRemove-1),
    rightSubarray = TNT::subarray (A, 0, nrows, colToRemove+1, ncols),
    colRemovedArray = TNT::cbind(leftSubarray, rightSubarray);

  return colRemovedArray;

  
}

/* write an array2D as a R/S-plus style matrix */
void TNT::writeArray(Array2D<double> &array, std::string filename)
{
  std::ofstream *dataStream = new std::ofstream(filename.c_str());
  if (dataStream->fail())
    {
      std::cerr << "Unable to open output file " << filename << std::endl;
      exit(1);
    }
  writeArray (array, *dataStream);  
  dataStream->close();
}

/* write an array2D as a R/S-plus style matrix */
void TNT::writeArray(Array2D<double> &array, std::ostream &out)
{
  int row, col,
    nRows = array.dim1(), nCols = array.dim2();

  // write out the data, each row as one line
  for (row = 0; row < nRows; row++){

    for (col = 0; col < nCols; col++){
      out << array[row][col];
      if (col < nCols-1) out << ' ';	// if not at end of line, put a space.
    }
    if (row < nRows -1) out << std::endl; // if not at end of file, put a newline.
  }
 }



// Read an array from a file in R/S format
Array2D <double> TNT::readDoubleArray2D(std::string filename)
{
  std::ifstream dataStream (filename.c_str());
  if (dataStream.fail())
    {
      std::cerr << "Unable to open input file " << filename << std::endl;
      exit(1);
    }
  Array2D <double> array = readDoubleArray2D (dataStream);  
  dataStream.close();

  return(array);
}

// Read one line of data from the input stream, and return it as a
// vector of doubles
vector <double> readDataLine (std::istream &input)
{
  vector <double> dataVec;
  double entry;

  while((input.peek() != '\n') & !input.eof())
    {
      if (!(input >> entry)) {
	cerr << "input entry is not a double" << endl;
	exit(1);
      }
      dataVec.push_back(entry);	
    }
  if (input.peek() == '\n') input.ignore(1);  //if we are at the end of a line,
  //move past the newline character	

  return (dataVec);
}

// Convert a vector of vector of doubles to an Array2D.  Each vector
// of doubles in vectorMatrix is assumed to be a row of the matrix.
// In other words, the input should be row-major.  Check that all of
// the rows have the same length.
Array2D <double> matrixToArray 
(const vector <vector<double> > & vectorMatrix){

  int nrow = vectorMatrix.size();
  if (nrow == 0) {
    return Array2D <double> ();
  }

  int ncol = vectorMatrix[0].size();

  vector<double> row;
  Array2D <double> data (nrow, ncol);  

  int i, j;
  for (i=0; i<nrow; i++){
    row = vectorMatrix[i];

    if ((int)row.size() != ncol){
      cerr << "inconsistent number of columns" << endl;
      exit(1);
    }

    for (j=0; j<ncol; j++)
    {
      data[i][j] = row[j];
    }
  }

  return (data);

}


// Read an array in R/S format from an input stream
Array2D <double> TNT::readDoubleArray2D(std::istream &in)
{
  vector<double > dataVector;
  vector<vector<double> > dataMatrix;

  while (!in.eof()){
    dataVector = readDataLine(in);
    if (dataVector.size() != 0)
      dataMatrix.push_back (dataVector);
    
  }

  return matrixToArray (dataMatrix);
}

// Read an array from a file in R/S format
Array2D <int> TNT::readIntArray2D(std::string filename)
{
  std::ifstream dataStream (filename.c_str());
  if (dataStream.fail())
    {
      std::cerr << "Unable to open input file " << filename << std::endl;
      exit(1);
    }
  Array2D <int> array = readIntArray2D (dataStream);  
  dataStream.close();

  return(array);
}

// Read one line of data from the input stream, and return it as a
// vector of doubles
vector <int> readIntDataLine (std::istream &input)
{
  vector <int> dataVec;
  int entry;

  while((input.peek() != '\n') & !input.eof())
    {
      if (!(input >> entry)) {
	cerr << "input entry is not a double" << endl;
	exit(1);
      }
      dataVec.push_back(entry);	
    }
  if (input.peek() == '\n') input.ignore(1);  //if we are at the end of a line,
  //move past the newline character	

  return (dataVec);
}

// Convert a vector of vector of ints to an Array2D.  Each vector
// of ints in vectorMatrix is assumed to be a row of the matrix.
// In other words, the input should be row-major.  Check that all of
// the rows have the same length.
Array2D <int> matrixToArray 
(const vector <vector<int> > & vectorMatrix){

  int nrow = vectorMatrix.size();
  if (nrow == 0) {
    return Array2D <int> ();
  }

  int ncol = vectorMatrix[0].size();

  vector<int> row;
  Array2D <int> data (nrow, ncol);  

  int i, j;
  for (i=0; i<nrow; i++){
    row = vectorMatrix[i];

    if ((int)row.size() != ncol){
      cerr << "inconsistent number of columns" << endl;
      exit(1);
    }

    for (j=0; j<ncol; j++)
    {
      data[i][j] = row[j];
    }
  }

  return (data);

}

// Read an array in R/S format from an input stream
Array2D <int> TNT::readIntArray2D(std::istream &in)
{
  vector<int > dataVector;
  vector<vector<int> > dataMatrix;

  while (!in.eof()){
    dataVector = readIntDataLine(in);
    if (dataVector.size() != 0)
      dataMatrix.push_back (dataVector);
  }

  return matrixToArray (dataMatrix);
}



/* Multiply a matrix and a vector */
Array1D <double> TNT::matmult (const Array2D<double> &A, const Array1D<double> &B){

  int i, len = B.dim();

  // Create a matrix with one column, which is equal to B
  Array2D<double> Bmat (len, 1);
  for (i=0; i<len; i++)
    Bmat[i][0] = B[i];

  Array2D<double> ABmatprod = matmult(A, Bmat);
  Array1D<double> ABprod (len);

  for (i=0; i<len; i++)
    ABprod[i] = ABmatprod[i][0];

  return(ABprod);
}

/* Calculate the sum of all of the elements */
double TNT::sum (const Array2D<double> &A){
  int i, j, nrow = A.dim1(), ncol = A.dim2();
  double total = 0.0;

  for( i=0; i<nrow; i++ )
    for( j=0; j<ncol; j++ )
      total += A[i][j];

  return( total );
}
